
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#include "myassertions.h"
#include "funcs.h"
#include "Eqn.h"

extern bool evolution_criterion(int times);

void Eqn::parse_generator(Generator const & g){

   PRECONDITION( (int) g.space_dimension() == _n+1 , "Wrong generator space dimension");
   
   int i;
   
   for (i=0;i<=_n;++i){
      Coefficient const & c = g.coefficient(Variable(i));

      if (sgn(c)==0)
         continue;

      PRECONDITION( c.fits_slong_p() , " Fatal Error: Coefficient cannot fit inside long ");

      long k= c.get_si();

      _m[i]=k;
      
   }

   return;
}

void Eqn::parse_constraint(Constraint const & cc, int scale){
   
   PRECONDITION( (int) cc.space_dimension() == _n , "Wrong generator space dimension");
   
   int i;
   
   for (i=0;i<_n;++i){
      Coefficient const & c = cc.coefficient(Variable(i));

      if (sgn(c)==0)
         continue;

      PRECONDITION( c.fits_slong_p() , " Fatal Error: Coefficient cannot fit inside long ");

      long k= scale * c.get_si();
      
      _m[i]=k;
      
   }

   Coefficient const & c = cc.inhomogeneous_term();
   
   if (sgn(c)==0)
      return;

   PRECONDITION( c.fits_slong_p() , " Fatal Error: Coefficient cannot fit inside long ");

   long k= scale * c.get_si();
      
   _m[_n]=k;
   

   return;
}

bool Eqn::is_evolving() const{  
   return !_frozen && evolution_criterion(_times);
}


long Eqn::operator() (int k) const{

   PRECONDITION( (k>=0 && k <= _n), " Out of range ");
   
   map<int,long>::const_iterator vi = _m.find(k);

   if (vi== _m.end())
      return 0;

   return vi->second;
   
}

long & Eqn::operator[](int k){

   PRECONDITION( (k>=0 && k <= _n), " Out of range ");
   
   map<int,long>::iterator vi = _m.find(k);

   if (vi== _m.end()) {
      pair<int,long> ipair(k,0); 
      vi= (_m.insert( ipair)).first;
   }

   return vi->second;
   
}



void Eqn::print(ostream & out) const{

   out<< "";
   for (int i=0;i<_n+1;++i){
      out<< operator() (i)<<"  ";
   }
   out<<" >= 0";

   return;
}


void Eqn::print_using(ostream & out, var_info * f) const{
   long k;
   out<< "";
   for (int i=0;i<_n;++i){
      k=operator()(i);
      if (k==0) continue;
      out << k<< " * "<< f->get_name(i)<<" + ";
   }
   
   k=operator()(_n);
   out << k<< " >=0";
   return;
}

ostream & operator<<( ostream & out, Eqn const & e){
   e.print(out);
   return out;
}


bool Eqn::positive_matches(Eqn const & what) const{
   int i;
   
   if (what._n !=_n) return false;
   
   int factord, factorn;
   

   for ( i=0; i < _n && what(i)==0 && this->operator()(i)==0; ++i) {}

   if (i >= _n) return true;

   if (what(i)==0 || this->operator()(i)==0)
      return false;

   factord = this->operator()(i);
   factorn= what(i);

   if  ( (factord <= 0 && factorn >=0) || (factorn <= 0 && factord >=0) )
      return false;
   
   ++i;
   for (;i<_n ;++i){
      if (factord * what(i)!= this->operator()(i) * factorn)
         return false;
   }

   return true;
                  
}

void Eqn::scale_up( int den){

  PRECONDITION( den>0, " Asked to scale by negative number, or zero. error somewhere?");
  
  if (den==1) return;

  map<int,long>::iterator vi;
  
  for (vi=_m.begin(); vi != _m.end() ; ++vi){
    long & what=vi->second;
    what *= den;
  }

  return;
}

bool Eqn::compare( Eqn const & c, long & l1, long & l2) const{
   int i;
   bool factored=false;
   for (i=0;i<_n;++i){
      long const & x = c(i);
      long const & y = operator()(i);

      if (x==0 && y==0) continue;
      
      if ( x ==0 && y!=0) return false;
      if ( x !=0 && y ==0) return false;

      if ( x >=0 && y < 0) return false;
      if ( x < 0 && y >=0) return false;

      
      if (!factored){
         long g= gcd(x,y);
         l1 = y/g;
         l2 = x/g;
         if (l1 < 0 && l2 < 0) {
            l1 = -l1;
            l2= -l2;
         }
            
         INVARIANT( l1 * x == l2 * y ,  " Invariant failure");
         INVARIANT(l1 >0 && l2 > 0, "Invariant failure");
         factored=true;
      } else{
         if (l1 * x != l2 *y) return false;
      }

      

      
      
   }

   if (!factored) return false;
   
   return true;

}

bool Eqn::is_zero() const{
   map<int,long>::const_iterator vi;
   for (vi=_m.begin(); vi!=_m.end(); ++vi){
      if (vi->second!=0)
         return false;
   }
   return true;
}

