
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#include "myassertions.h"
#include "EqnUtilities.h"

#define NOT_FEASIBLE 0
#define UNBOUNDED -1

extern bool perform_pre_optimization;

void buildEquationTemplate( GenPool const & gp, EquationTemplate & result){
   //
   // From each evolving gen, obtain the generator and add it to the template
   //

   int n = gp.obtain_dimensions();
   
   vector<EvolvingGens> const & vg = gp.obtain_generators();
   vector<EvolvingGens>::const_iterator vi;

   for (vi= vg.begin(); vi != vg.end(); ++vi){
      Generator const & g = vi->obtain_generator();
      Eqn temp(n,g);
      result.add_equation(temp);
   }
}

void buildEquationTemplate(C_Polyhedron const & p, EquationTemplate & result){

   int n= result.get_dimension();
   
   PRECONDITION( (int) p.space_dimension() == n , \
                 " Space dimensions incompatible");

   Constraint_System cs = p.constraints();
   Constraint_System::const_iterator vi;
   for (vi = cs.begin(); vi != cs.end(); ++vi){
      Eqn eq(n,(*vi));
      result.add_equation(eq);
      if (vi->is_equality()){
         Eqn eq1(n,(*vi),-1);
         result.add_equation_blindly(eq1);
      }
         
   }

   return;
   
}

void buildLProblem(C_Polyhedron const & p, LProblem & lpr){
   /*
    * 1. First count the required number of rows and columns
    * 2. Then start writing the LProblem
    */
   //   PRECONDITION( !p.is_empty(), " buildLProblem not designed to handle empty polyhedra");
   
   int n = (int)p.space_dimension();
   

   //
   // count the number of column variables
   //

   Constraint_System cs = p.constraints();
   Constraint_System::const_iterator vi;

   int ncols=0,i=0;

   for (vi=cs.begin(); vi != cs.end(); ++vi, ++ncols);

   // form a constraint with n+1 rows and ncols+1 cols

   lpr.reset( n+1 , ncols+1);

   // make the last column positive and
   // add a +1 >=0 constraint
   lpr.make_col_positive(ncols);
   lpr.set_coefficient(n,ncols,1);
   lpr.set_objective_col(ncols,1);
   i=0;
   
   for (vi=cs.begin(); vi != cs.end(); ++vi, ++i){
      handle_constraint((*vi),i,n, lpr);
   }

   // now set an objective
   // and set the problem to a minimization

   
   lpr.set_maximize(false);
   // lpr.set_objective_row(n, 1);
   
   // objectives set too

   lpr.load_problem();
   
}







void buildLProblem(EquationTemplate const & pre, C_Polyhedron const & guard,  LProblem & result){

   /*
    * 1. First count the required number of rows and columns
    * 2. Then start writing the LProblem
    */

   int n = pre.get_dimension();
   PRECONDITION( (int) guard.space_dimension() == n , " Wrong dimensions for the guard ");

   //
   // count the number of column variables
   //

   Constraint_System cs = guard.constraints();
   Constraint_System::const_iterator vi;
   int ncols_g=0,i=0;

   for (vi=cs.begin(); vi != cs.end(); ++vi, ++ncols_g);

   
   int ncols_e = pre.get_num_equations();

   //
   // the problem result will have n+1 rows and
   // ncols_e+_g columns
   //
   
   result.reset( n+1, ncols_e+ ncols_g+1);

   // add the column for 1 >= 0 reqd. by Farkas' Lemma
   
   result.make_col_positive(ncols_e+ncols_g);

   result.set_coefficient(n,ncols_e+ncols_g, 1);
   result.set_objective_col(ncols_e+ncols_g,1);
   //
   // now make all the columns corresponding the the EquationTemplate,
   // positive to start with
   //

   for (i=0;i<ncols_e;++i){
      result.make_col_positive(i);
   }

   i=0;
   
   for (vi=cs.begin(); vi != cs.end(); ++vi,++i){

      //
      // now if this row is an equality, then leave it free
      // else add a positivity constraint
      //
      
      handle_constraint((*vi), ncols_e+i,n, result);
   }

   // now iterate across the equations and add_constraints

   EquationTemplate::const_iterator vj;

   i=0;
   
   for (vj = pre.begin(); vj != pre.end(); ++vj, ++i){
      //
      // add the coefficients for the equation
      //
      
      handle_eqn( (*vj), i, n, result); 
   }

   
   // 
   // all bounds have been set for the columns,
   // all coefficients have been loaded
   // 

   result.set_maximize(false);

   // disabled... GLPK no longer handles row objective coefficients
   
   //result.set_objective_row(n, 1);
   
   // objectives set too

   result.load_problem();

}

void buildLProblem(EquationTemplate const & pre, LProblem & result){

   PRECONDITION( !pre.inconsistent(), " Asked to build LProblem for an inconsistnet set of inequalities");
   int n = pre.get_dimension();

   C_Polyhedron dummy(n,UNIVERSE);

   buildLProblem(pre,dummy,result);

   return;
}

void handle_constraint( Constraint const & cc, int col_num, int n, LProblem & result){

   PRECONDITION(cc.is_nonstrict_inequality() || cc.is_equality(), " Strict constraints not handled ");

   if (cc.is_nonstrict_inequality()){
      result.make_col_positive(col_num);
   }
   
   int i;

   for (i=0 ; i < n ; ++i){
      Coefficient const & coeff = cc.coefficient(Variable(i));

      if (sgn(coeff)==0) continue;
      
      PRECONDITION ( coeff.fits_slong_p(), " Overflow error ");

      long k = coeff.get_si();
      result.set_coefficient(i,col_num, k);
   }

   Coefficient const & coeff = cc.inhomogeneous_term();
   
   if (sgn(coeff) !=0) {
      
      PRECONDITION ( coeff.fits_slong_p(), " Overflow error ");
      result.set_coefficient(n,col_num, coeff.get_si());
      // also set this to be the coefficient of the objectice column col_num
      result.set_objective_col(col_num,coeff.get_si());
   }

   return;
   
}


void handle_eqn(Eqn const & eq, int col_num, int n, LProblem & result){

   int i;
   
   for (i=0;i<=n;++i){
      result.set_coefficient(i, col_num, eq(i));
      if (i==n)
         result.set_objective_col(col_num,eq(i));
   }

   return;
}

/*
 * Function: post_computation
 *   Computes the post-condition for the equation templates pre and post w.r.t
 *   a given transition trs
 */


void post_computation( EquationTemplate & pre, EquationTemplate & post, TrRel const & trs){
   // int n = pre.get_dimension();
   
   //
   // if precondition is inconsistent then
   // return;
   //
   
   if (pre.inconsistent())
      return;

   //
   // else choose each eqn inside the post condition and compute a "better" value
   //

   // first compute a LProblem

   LProblem pr(1,1);

   //
   // build the problem for the pre /\ guard
   //

   // asking for transition guard
  
   buildLProblem(pre, trs.get_guard(), pr);
  
   //
   // build the problem for the guard along just to
   // make local widening easier for later. This done lazily.
   //

   bool local_widen_lproblem_built=false;
   
   LProblem pr1(1,1);


   
   
   EquationTemplate::iterator vi;



   for (vi=post.begin(); vi != post.end(); ++vi){
      Eqn & c = (*vi);

      //
      // frozen templates have been widened to as far as they
      // can reach. So no point updating them.
      //
      
      //if (c.is_frozen()) continue;
      
      pair<long, long> res(0,0);
      bool plb;

      
      
      if (c.is_evolving()){
         
         plb = optimize_coefficient_fractional(pre, c, trs, pr, res,perform_pre_optimization);
      
         if (plb){
	
            bool flag=post.update_constant(vi, res.first,res.second);
            if (flag)
               --vi;
	
         } else {

            if (res.first==NOT_FEASIBLE){
               vi=post.drop_equation(vi);
               vi--;
            }

            if (res.first == UNBOUNDED){
               // then can the post computation be aborted?
               return;
            }
            
         }
      } else {

         //
         // Local invariant computation
         // takes over here
         //

         // if the LProblem for local widening has never been built,
         // build it here

         if (!local_widen_lproblem_built){
            buildLProblem(trs.get_guard(), pr1);
            local_widen_lproblem_built=true;
         }

         // optimize based on this lproblem
         plb=optimize_coefficient_fractional(pre,c,trs,pr1,res);

         // check the status of the solution
         if (plb){

            //
            // optimal solution found
            // then update it but make sure that
            // the evolution criteria is not used
            //
            
            post.update_constant_frozen(vi, res.first, res.second);  
         } else {
            // if not feasible,
            // then return
            
            if (res.first==NOT_FEASIBLE){
               vi=post.drop_equation(vi);
               vi--;
            }

            if (res.first==UNBOUNDED){
               // guard unsatisfiable
               // abort the computation
               return;
            }


         }
         
      }
   }
   
   if (post.inconsistent()){
      post.set_inconsistent(false);    
   }
   
   
}

long handle_with_exception(Coefficient const & c){
   PRECONDITION( c.fits_slong_p(), " Coefficient overflow error  ");

   return c.get_si();
}


long evaluate(Linear_Expression const & ll, Eqn const & what){

   int n= what.get_dimension();
   
   Coefficient result;

   result=0;
   int j;
  
   for (j=0; j<=n; ++j){
      if (what(j)==0) continue;
      result +=  what(j) * ll.coefficient(Variable(j));
   }
   
   return handle_with_exception(result);
   
}

void update_map( int n, Eqn const & what, TrRel const & trs, Eqn & result){

   int i;

   for (i=0; i<=n; ++i){
      Linear_Expression const & ll = trs.get_update_expression(i);
      result[i]= evaluate(ll, what);
   }
}



pair<long,bool> optimize_coefficient(EquationTemplate & pre, Eqn & c, TrRel const & trs, LProblem & lpr){

   PRECONDITION( ! pre.inconsistent(),  " Asked to optimize for {false} -- not possible ");

   
   int n = pre.get_dimension();

   long r=c(n);
   
   c[n]=0;

   //
   // now first compute the updated coefficients
   // the value of c(n) is ignored and treated as
   // zero by the update_map function
   // therefore c1(n) shows the offset value that
   // must be subtracted from the optimal value
   // returned
   //

   Eqn c1(n);
   
   update_map ( n, c, trs, c1);
   
   
   c[n]=r;
   
   

   return optimize_coefficient(c1,lpr, c1(n));
}



bool optimize_coefficient_fractional (EquationTemplate & pre, Eqn & c, TrRel const & trs, \
				      LProblem & lpr, pair<long, long> & respair, bool preopt){

   PRECONDITION( ! pre.inconsistent(),  " Asked to optimize for {false} -- not possible ");

   
   int n = pre.get_dimension();

   long r=c(n);
   
   c[n]=0;

   //
   // now first compute the updated coefficients
   // the value of c(n) is ignored and treated as
   // zero by the update_map function
   // therefore c1(n) shows the offset value that
   // must be subtracted from the optimal value
   // returned
   //

   Eqn c1(n);
   
   update_map ( n, c, trs, c1);
   
   
   c[n]=r;
   
   if (preopt && quick_optimize_fractional(c1, pre,respair)){

      long & num=respair.first;
      long & den=respair.second;
      num -= den * c1(n);
      return true;
   }
   
   return optimize_coefficient_fractional(c1,lpr, c1(n),respair);
}


bool quick_optimize_fractional(Eqn &c, EquationTemplate & pre, pair<long,long> & result){
   long a,b;
   
   if (pre.compute_best_positive_value(c,a,b)){
      result.first=a;
      result.second=b;
      return true;
   }

   return false;
}

bool optimize_coefficient_fractional( Eqn & c, LProblem & lpr, long offset, pair<long,long> & result){

   int n=c.get_dimension();

   int i;

   for (i=0;i<n;++i){

      //
      // set row $i$s upper and lower bounds to be c1(i)
      //

      lpr.set_row_upper_bound(i,c(i));
      lpr.set_row_lower_bound(i,c(i));
      
   }

   
   lpr.rebuild_bounds_rows();

     
   lpr.solve();
   
   if (lpr.interpret_status()){
      // all is well
     pair<long,long> temp = lpr.get_optimal_value_fractional();
     long num=temp.first;
     long den = temp.second;
     num  -= offset * den;
     
     result.first=num;
     result.second = den;

     return true;
   } else if (!lpr.is_feasible()){
      // not feasible
     result.first=NOT_FEASIBLE;
     return false;
     
      
   } else if (lpr.is_unbounded()){

      // unbounded..  guard \intersect assertion is probably false
     result.first=UNBOUNDED;
     return false;

   } else{
      // solver trouble?
      UNREACHABLE(" Unknown LP solver error");
      exit(1);
   }
   
   
   
    result.first=NOT_FEASIBLE;
    return false;
    

}


pair<long,bool> optimize_coefficient( Eqn & c, LProblem & lpr, long offset){
   
   int n=c.get_dimension();
   int i;

   for (i=0;i<n;++i){

      //
      // set row $i$s upper and lower bounds to be c1(i)
      //

      lpr.set_row_upper_bound(i,c(i));
      lpr.set_row_lower_bound(i,c(i));
      
   }

   
   lpr.rebuild_bounds_rows();

     
   lpr.solve();
   if (lpr.interpret_status()){
      // all is well
      
      return pair<long,bool>(lpr.get_optimal_value_truncated() - offset, true);
      
   } else if (!lpr.is_feasible()){
      // not feasible
      return pair<long,bool>(0,false);
      
   } else if (lpr.is_unbounded()){

      // unbounded..  guard \intersect assertion is probably false
      return pair<long,bool>(-1 ,false);

   } else{
      // solver trouble?
      UNREACHABLE(" Unknown LP solver error");
      exit(1);
   }
   
   return pair<long,bool>(0,false);


}



void optimize_and_add(Eqn & c, EquationTemplate & what , LProblem & lpr, bool blind){

   int n= what.get_dimension();
   
    pair<long,long> res(0,0);
    bool status = optimize_coefficient_fractional(c, lpr, 0, res);
   
   if (status){
     long num=res.first;
     long den=res.second;

     // scale up the equation c
     c. scale_up(den);
     c[n]=num;

     if (!blind)
        what.add_equation(c);
     else
        what.add_equation_blindly(c);
     
   } else {
     if (res.first == UNBOUNDED){

        /*
         * If this happens, then the underlying polyhedron is
         * empty, but I am handling that separately. Why then am
         * I reaching this point?
         */ 
        cerr<< " WARNING -- underlying polyhedron is empty?"<<endl;
        cerr<< " I am not supposed to reach this under normal operation.. check the results for invariance :-)"<<endl;
     }
   }  
}


void create_bounded_equations( EquationTemplate & what, LProblem & lpr){
   int n=what.get_dimension();
   Eqn c(n);
   for (int i=0; i< n ; ++i){
      c[i]=1;
      optimize_and_add(c,what,lpr);
      c[i]=-1;
      optimize_and_add(c,what,lpr);
      c[i]=0;
   }

}

void create_octagons(EquationTemplate & what, LProblem & lpr){
   int n=what.get_dimension();
   Eqn c(n);
   int i,j;
   for ( i=0; i< n-1 ; ++i){
      c[i]=1;
      for(j=i+1;j< n ; ++j){
         c[j]=-1;
         optimize_and_add(c,what,lpr);
         c[j]=1;
         optimize_and_add(c,what,lpr);
         c[j]=0;
      }

      c[i]=-1;
      for(j=i+1;j< n ; ++j){
         c[j]=-1;
         optimize_and_add(c,what,lpr);
         c[j]=1;
         optimize_and_add(c,what,lpr);
         c[j]=0;
      }

      c[i]=0;
   }

}


void create_bounded_equations(EquationTemplate & what){
   int n=what.get_dimension();
   Eqn c(n);
   for (int i=0; i< n ; ++i){
      c[i]=1;
      what.add_equation(c);
      c[i]=-1;
      what.add_equation(c);

      c[i]=0;
   }

}


void create_octagons(EquationTemplate & what){
   int n=what.get_dimension();
   Eqn c(n);
   int i,j;
   for ( i=0; i< n-1 ; ++i){
      c[i]=1;
      for(j=i+1;j< n ; ++j){
         c[j]=-1;
         what.add_equation(c);
         c[j]=1;
         what.add_equation(c);
         c[j]=0;
      }

      c[i]=-1;
      for(j=i+1;j< n ; ++j){
         c[j]=-1;
         what.add_equation(c);
         c[j]=1;
         what.add_equation(c);
         c[j]=0;
      }
      
      c[i]=0;
   }

}



void add_extra_relations(EquationTemplate & what, LProblem & lpr){
   create_bounded_equations(what,lpr);
   create_octagons(what,lpr);

}


void add_extra_relations(EquationTemplate & what){
   create_bounded_equations(what);
   create_octagons(what);
}

void enrich_pair(EquationTemplate & pre, EquationTemplate & post, TrRel const & trs, LProblem & lpr){

   PRECONDITION( !pre.inconsistent(), "  Error:Suspicious LProblem built for an inconsitent set of equations");

   
   int n=pre.get_dimension();
   EquationTemplate::const_iterator vi;

   for (vi=post.begin(); vi != post.end(); ++vi){
      Eqn p(n);
      update_map(n,(*vi), trs, p);
      if (!pre.has_equation(p))
         optimize_and_add(p,pre,lpr, true);
   }
   return;
}


void enrich_pair(EquationTemplate & pre, EquationTemplate & post, TrRel const & trs){

   PRECONDITION( pre.inconsistent(), "  Error: Function called minus the LProblem");
   
   int n=pre.get_dimension();
   
   EquationTemplate::const_iterator vi;

   for (vi=post.begin(); vi != post.end(); ++vi){
      Eqn p(n);
      update_map(n,(*vi), trs, p);
      pre.add_equation(p);
   }
   return;
}


void merge_templates( EquationTemplate & what, EquationTemplate const & from){
   EquationTemplate::const_iterator vi;
   for (vi=from.begin(); vi!=from.end(); ++vi){
      what.add_equation((*vi));
   }
}


void enrich_pair(EquationTemplate & pre,  TrRel const & trs, LProblem & lpr){

   PRECONDITION( !pre.inconsistent(), "  Error:Suspicious LProblem built for an inconsitent set of equations");

   
   int n=pre.get_dimension();
   EquationTemplate dummy(n);
   
   EquationTemplate::const_iterator vi;

   for (vi=pre.begin(); vi != pre.end(); ++vi){
      Eqn p(n);
      update_map(n,(*vi), trs, p);
      if (!pre.has_equation(p))
         optimize_and_add(p,dummy,lpr,true);
   }
   merge_templates(pre,dummy);
   return;
}


void enrich_pair(EquationTemplate & pre, TrRel const & trs){

   PRECONDITION( pre.inconsistent(), "  Error: Function called minus the LProblem");
   
   int n=pre.get_dimension();

   EquationTemplate dummy(n);
   
   EquationTemplate::const_iterator vi;

   for (vi=pre.begin(); vi != pre.end(); ++vi){
      Eqn p(n);
      update_map(n,(*vi), trs, p);
      dummy.add_equation(p);
   }

   merge_templates( pre, dummy);
   
   return;
}


