
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */



#include "myassertions.h"
#include "EquationTemplate.h"


extern int current_iteration;

bool EquationTemplate::has_equation(Eqn const & eq){

   if (eq.get_dimension()!=_n) return false;

   
   iterator vi;
   
   for (vi=_ve.begin(); vi!= _ve.end(); ++vi){
      if (vi->positive_matches(eq))
         return true;
   }

   return false;
}

void EquationTemplate::add_equation(Eqn const & eq){

   PRECONDITION( eq.get_dimension() == _n, " Asked to add equation with incompatible dimensions");

   if (has_equation(eq))
      return;
   
   _ve.push_back(eq);

}


void EquationTemplate::add_equation_blindly(Eqn const & eq){

   PRECONDITION( eq.get_dimension() == _n, " Asked to add equation with incompatible dimensions");

   
   _ve.push_back(eq);

}


EquationTemplate::iterator EquationTemplate::drop_equation(EquationTemplate::iterator const & vi){
   _iter=current_iteration;
   return _ve.erase(vi);
}




bool EquationTemplate::update_constant( EquationTemplate::iterator & vi, long j, long d){
   
   // flag true if equation has to drop the generator

   if (_incon){
      
      // then just set it to whatever it has to be
      Eqn & c= (*vi);
      c.scale_up(d);
      c[_n]=j;
      _iter=current_iteration;
      return false;
   }
   
   if (!vi->is_evolving()){
      vi=drop_equation(vi);
     _iter=current_iteration;
     return true;
   } else {
     Eqn & c= (*vi);
     if (c[_n] * d < j){
        c.scale_up(d);
        c[_n]=j;
        c.increase_times();
        _iter=current_iteration;
     }
   }

   return false;
}


bool EquationTemplate::update_constant_frozen( EquationTemplate::iterator & vi, long j, long d){
   // if frozen, then make _frozen true
   vi->freeze(true);
   // flag true if equation has to drop the generator
   
   if (_incon){
      // then just set it to whatever it has to be
      Eqn & c= (*vi);
      c.scale_up(d);
      c[_n]=j;
      _iter=current_iteration;
      return false;
   }
   
   
   
   
   Eqn & c= (*vi);
   
   if (c[_n] * d < j){
      _iter=current_iteration;
      c.scale_up(d);
      c[_n]=j;
      c.increase_times();
   }
   
   return false;
}

void EquationTemplate::print(ostream & out) const{

   vector<Eqn>::const_iterator vi;
   for (vi=_ve.begin(); vi != _ve.end(); ++vi){
      out<<(*vi)<<endl;
   }
   return;
}


ostream & operator<<(ostream & out, EquationTemplate const & eqt){
   eqt.print(out);
   return out;
}

void EquationTemplate::print_using(ostream & out, var_info * f) const{
   vector<Eqn>::const_iterator vi;
   if (_incon) {
      cerr<<"(**)";
   }
   cerr<<" Iteration :"<<_iter<<endl;
   for (vi=_ve.begin(); vi != _ve.end(); ++vi){
      vi->print_using(out,f);
      out<<endl;
   }
   return;
}


void EquationTemplate::convert(C_Polyhedron & what) const{

   if (_incon){
      what=C_Polyhedron(_n, EMPTY);
      return;
   }
   //what = C_Polyhedron(_n, UNIVERSE);
   const_iterator vi;
   int i;
   
   for (vi=_ve.begin(); vi != _ve.end(); ++vi){
      Eqn const & e=(*vi);
      Linear_Expression ll(e(_n));
      
      
      for (i=0;i<_n;++i)
         ll+= e(i) * Variable(i);
      what.add_constraint(ll >= 0);
   }
   
}

EquationTemplate::const_iterator EquationTemplate::search_equation(Eqn const & c, \
                                                 long & l1, \
                                                 long & l2){
   const_iterator vi;

   for (vi=this->begin(); vi!= this-> end(); ++vi){

      if (vi->compare(c,l1,l2)){
         return vi;
      }

   }

   return this->end();

}

bool EquationTemplate::compute_best_positive_value(Eqn const & c, long & a, long & b){
   
   
   
   long l2;
   const_iterator vi = search_equation(c,b,l2);
   if (vi==this->end()) return false;
   
   a=l2 * vi->operator()(_n);

   return true;
   
}
