
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#ifndef D__EQUATION__TEMPLATE__H_
#define D__EQUATION__TEMPLATE__H_

#include <iostream>
#include <vector>
#include "var-info.h"
#include "ppl.hh"
#include "Eqn.h"
#include "GenPool.h"
#include "LProblem.h"

using namespace std; 
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;

/*
 * An equation template is of the form c_i = \sum_j=1^m a_{ij} \lambda_j 
 *   each $\lambda_ij >=0$
 *
 *   Operations supported are
 *      1. From a generator pool, build an equation template,
 *      2. Drop an equation.
 *     
 */


class EquationTemplate{
  private:
   
   vector<Eqn> _ve;
   int _n;
   bool _incon;
   int _iter;
   
   
  public:

   
   typedef vector<Eqn>::iterator iterator;
   
   typedef vector<Eqn>::const_iterator const_iterator;
   
   EquationTemplate(int n):_n(n),_incon(false), _iter(0){}

   void add_equation(Eqn const & eq);
   void add_equation_blindly(Eqn const & eq);
   
   bool inconsistent() const{
      return _incon;
   }

   void set_inconsistent(bool incon=true){
      _incon=incon;
     
   }
   
   iterator begin(){
      return _ve.begin();
   }
   iterator end(){
      return _ve.end();
   }

   
   const_iterator begin() const{
      return _ve.begin();
   }
   const_iterator end() const{
      return _ve.end();
   }

   
   iterator drop_equation(iterator const & vi); // iterator will no longer be valid
   
   int get_dimension() const{
      return _n;
   }

   int get_num_equations() const{
      return (int) _ve.size();
   }

   void convert(C_Polyhedron & what) const;
   
   // update constant coefficient to c
   // and multiply the rest of the equation by d
   // if return value is true then the
   // generator is dropped.
   
   bool update_constant( iterator & vi, long c, long d=1);
   bool update_constant_frozen( iterator & vi, long c, long d=1);

   bool has_equation( Eqn const & p);
   
   void print(ostream & out) const;
   void print_using(ostream & out, var_info * f ) const;

   
   int iteration_count() const{
      return _iter;
   }

   /*  check if there is a Eqn e in this template s.t. for some multipliers
    *        l1,l2 > 0   l1. c = l2. e and for the constant term,
    * 
    */

   const_iterator search_equation(Eqn const & c,long & l1, long & l2);
   
   /*
    * call search equation to find a suitable e s.t. l1 * c = l2 * e
    * return a = l2 * e[n]
    *        b = l1
    */
   
   bool compute_best_positive_value(Eqn const & c, long & a, long & b);
   
   
};

ostream & operator<<(ostream & out, EquationTemplate const & eqt);



#endif
