
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include "EvolvingGens.h"
#include "OneParameterConstraint.h"

//
// the maximum number of evolutions
// before the generator gets dropped.
//

extern int _max_evolutions;

void EvolvingGens::adjust_constant_coefficient( C_Polyhedron const & ref ){

   //
   // for each constraint in ref,
   // plug in generator g for variables 0.. n-1.
   // if variable n is zero.. then check if the value is positive/zero
   // if variable n is not zero, then maintain a maximum so far
   //

   if (!_valid)
      return;
   if (_times > _max_evolutions){
      _valid=false;
      return;
   }

   OneParameterConstraint s;
   
   int i;
   
   Constraint_System cs=ref.constraints();
   Constraint_System::const_iterator vi;

   // iterate through the constraints
   
   for (vi=cs.begin(); vi!= cs.end(); ++vi){
      Constraint const & cc=(*vi);
      int sum=0;

      //
      // compute sum = (constraint). (g) for the first
      // 0..n-1 variables
      //
      
      for ( i=0;i<_n;++i){
         sum+= handle_integers(cc.coefficient(Variable(i)) * _g.coefficient(Variable(i))); 
      }
      
      if (sgn(cc.coefficient(Variable(_n)))== 0) {

         //
         // if nth coefficient is zero
         // then check if sum makes sense
         //

         if (cc.is_nonstrict_inequality() && sum < 0 ) {
            _valid=false;
            return;
         } else if (cc.is_equality() && sum != 0) {
            _valid=false;
            return;
         } else if(cc.is_strict_inequality() && sum <= 0) {
            _valid=false;
            return;
         }
         
      } else {
         
         //
         // now compute what the constraint says d must be
         // greater than
         //
         
         Rational t( -sum, handle_integers(cc.coefficient(Variable(_n))));
         
         if (cc.is_equality())
             s.eq(t);
         else if(cc.is_nonstrict_inequality())
            s.geq(t);
         else
            s.ge(t);

         if (s.is_empty()) {
            _valid=false;
            return;
         }
            
      }

   }
      
   if (s.is_empty()){
      _valid=false;
      return;
   }
      
   INVARIANT(! s.is_universal(),
             " Polyhedron ref seems to have an empty primal ");
   INVARIANT( s.has_lower_limit(),
              " Funny I always thought this should hold here ");

   
   Rational t = s.lower_limit();
   int nu=t.num() , de= t.den();
   
   Linear_Expression ll;
   for (i=0;i<_n;++i){
      ll+=  de * ( _g.coefficient(Variable(i)) * Variable(i));
   }

   ll+= nu * Variable(_n);
   
   if (_g.is_ray())
      _g=ray(ll);
   else if (_g.is_line())
      _g=line(ll);
   else if (_g.is_point())
      _g=point(ll);
   else if (_g.is_closure_point())
      _g=closure_point(ll);
   
   _times++;
}


EvolvingGens::EvolvingGens(int n, Generator const & g):
      _n(n),
      _g(g),
      _valid(true),
      _times(0){
   PRECONDITION ( _g.space_dimension() == _n+1, " Wrong space dimension");
}

EvolvingGens::EvolvingGens(int n, Generator const & g, C_Polyhedron const & ref):
      _n(n),
      _g(g),
      _valid(true),
      _times(-1){

   // discover the right constant to plug in if one such constant exists.
   
   adjust_constant_coefficient(ref);
   POSTCONDITION ( _g.space_dimension() == _n+1, " Wrong space dimension");
}



void EvolvingGens::print(ostream & out) const{
   out <<"<<";
   int i;
   for (i=0;i<_n+1;++i){
      out << _g.coefficient(Variable(i))<< "  ";
   }
   out<<">>"<<endl;
}

ostream & operator<< (ostream & out, EvolvingGens const & evg){
   evg.print(out);
   return out;
}

/* old code
   
// void EvolvingGens::adjust_constant_coefficient( C_Polyhedron const & ref ){

//    //
//    // for each constraint in ref,
//    // plug in generator g for variables 0.. n-1.
//    // if variable n is zero.. then check if the value is positive/zero
//    // if variable n is not zero, then maintain a maximum so far
//    //

//    if (!_valid)
//       return;
//    if (_times > _max_evolutions){
//       _valid=false;
//       return;
//    }

//    OneParameterConstraint s;
   
//    int i;
   
//    Constraint_System cs=ref.constraints();
//    Constraint_System::const_iterator vi;

//    // iterate through the constraints
   
//    for (vi=cs.begin(); vi!= cs.end(); ++vi){
//       Constraint const & cc=(*vi);
//       int sum=0;

//       //
//       // compute sum = (constraint). (g) for the first
//       // 0..n-1 variables
//       //
      
//       for ( i=0;i<_n;++i){
//          sum+= handle_integers(cc.coefficient(Variable(i)) * _g.coefficient(Variable(i))); 
//       }
      
//       if (sgn(cc.coefficient(Variable(_n)))== 0) {

//          //
//          // if nth coefficient is zero
//          // then check if sum makes sense
//          //

//          if (cc.is_nonstrict_inequality() && sum < 0 ) {
//             _valid=false;
//             return;
//          } else if (cc.is_equality() && sum != 0) {
//             _valid=false;
//             return;
//          } else if(cc.is_strict_inequality() && sum <= 0) {
//             _valid=false;
//             return;
//          }
         
//       } else if (!fixed && !cc.is_equality()) {
         
//          //
//          // now compute what the constraint says d must be
//          // greater than
//          //
         
//          Rational t( -sum, handle_integers(cc.coefficient(Variable(_n))));
//          if (!set){
//             set = true;
//             s=t;
//          } else  if ( t >= s) {
//             s=t;
//          }
         
//       } else if (!fixed && cc.is_equality()){

//          //
//          // if equality is encountered then
//          // value of s has been permanently fixed
//          // other constraints need to respect this
//          //
         
//          Rational t( -sum, handle_integers(cc.coefficient(Variable(_n))));
//          if (!set){
//             set = true;
//             s=t;
//             fixed=true;
//          } else  if ( t >= s) {
//             s=t;
//             fixed=true;
//          } else {
//             // the constraint system does not admit the generator
//             _valid=false;
//             return;
//          }
//       } else if (fixed){
//          INVARIANT(set, " Invariant violated.. unset constant has already been fixed ");
//          Rational t( -sum, handle_integers(cc.coefficient(Variable(_n))));

//          if (cc.is_equality() && !(t==s)) {
//             _valid=false;
//             return;
//          } else if (cc.is_nonstrict_inequality() && ( s < t )){
//             _valid=false;
//             return;
//          } else if (cc.is_strict_inequality() && (s <= t)){
//             _valid = false;
//             return;
//          }
//       }
      
//    }
   
//    INVARIANT(set, " Polyhedron ref seems to have an empty primal ");
   
//    int nu=s.num(), de= s.den();
//    Linear_Expression ll;
//    for (i=0;i<_n;++i){
//       ll+=  de * ( _g.coefficient(Variable(i)) * Variable(i));
//    }

//    ll+= nu * Variable(_n);

//    if (_g.is_ray())
//       _g=ray(ll);
//    else if (_g.is_line())
//       _g=line(ll);
//    else if (_g.is_point())
//       _g=point(ll);
//    else if (_g.is_closure_point())
//       _g=closure_point(ll);
   
//    _times++;
// }

*/
