
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include "GenPool.h"

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;

void GenPool::add_evolving_gen(EvolvingGens const & eg){
   if (eg.is_valid())
      _vg.push_back(eg);
}

void GenPool::initialize_vg_with(C_Polyhedron const & init){
   // first add all the generators of init itself

   Generator_System const & gs = init.generators();

   Generator_System::const_iterator vi;
   
   int i;
   
   for (vi=gs.begin(); vi != gs.end(); ++vi){
      if ((*vi).is_line()){
         Linear_Expression lin(0);
         for (i=0;i<=_n;++i)
            lin+= (*vi).coefficient(Variable(i)) * Variable(i);
         
         EvolvingGens eg1 (_n, ray(lin));
         add_evolving_gen(eg1);
         EvolvingGens eg2 (_n, ray(0-lin));
         add_evolving_gen(eg2);
      } else {
         EvolvingGens eg3 (_n, (*vi));
         add_evolving_gen(eg3);
      }
   }

   // now add some more generators whose limits need to be worked out

   for (i=0;i<_n;++i){
      Generator g1 = ray(Variable(i) + Variable(_n));
      Generator g2 = ray(- Variable(i) - Variable(_n));
      EvolvingGens eg4(_n, g1, init);
      add_evolving_gen(eg4);
      
      EvolvingGens eg5(_n, g2, init);
      add_evolving_gen(eg5);
   }

   // done

   return;
   
}

void GenPool::compute_polyhedron(){
   //
   // obtain the generators and add them to a polyhedron
   //

   _poly=C_Polyhedron(_n+1, UNIVERSE);
   int i;
   for (i=0;i<=_n;++i)
      _poly.add_constraint(Variable(i)==0);

   vector<EvolvingGens>::iterator vi;

   for (vi=_vg.begin(); vi != _vg.end(); ++vi){
      //
      // Obtain the generator from vi
      //

      if (!vi->is_valid()){
         vi=_vg.erase(vi);
         --vi;
         continue;
      }

      Generator & g=vi->obtain_generator();

      _poly.add_generator(g);
   }


   return;
   
}


void GenPool::narrowing_assign(C_Polyhedron const & next){
   // check each generator and recompute _poly

   if (_poly.is_universe()){

      if (next.is_universe()){
         return;
      }
      _poly=next;
      initialize_vg_with(next);
      return;
   }
   
   _poly=C_Polyhedron(_n+1, UNIVERSE);
   int i;
   for (i=0;i<=_n;++i)
      _poly.add_constraint(Variable(i)==0);

   vector<EvolvingGens>::iterator vi;

   for (vi=_vg.begin(); vi != _vg.end(); ++vi){
      //
      // Obtain the generator from vi
      //

      if (!vi->is_valid()){
         vi=_vg.erase(vi);
         --vi;
         continue;
      }

      Generator & g=vi->obtain_generator();
      if (next.relation_with(g) != Poly_Gen_Relation::subsumes()){
         vi->refine_with(next);
         if (!vi->is_valid()) {
            vi=_vg.erase(vi);
            --vi;
            continue;
         }
         _poly.add_generator(vi->obtain_generator());
      } else {
         _poly.add_generator(g);
      }
      
   }


   return;
}

void GenPool::print(ostream & out) const{

   if (_poly.is_universe()){
      out <<" True "<<endl;
      return;
   }
   
   vector<EvolvingGens>::const_iterator vi;

   for (vi=_vg.begin(); vi != _vg.end(); ++vi){
      out<<(*vi)<<endl;
   }
   return;
}


ostream & operator<<(ostream & out, GenPool const & gp){
   gp.print(out);
   return out;
}


