
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include "IMap.h"

IMap::IMap(vector<TrLoc> const & vloc, bool dual):_vloc(vloc){
   _dim=0;
   // initialize the map _m and add entries to it.

   vector<TrLoc>::const_iterator vi;
   
   for (vi = vloc.begin(); vi != vloc.end(); ++vi){
      string const & name= vi->get_name();
      _dim=vi->get_dimension();
      if (dual){
         _m.insert(StrPolyPair(name,vi->get_dinit()));
      
         GenVector v;
         add_initial_generators(v);
         _frames.insert(StrFramePair(name,v ));
      } else {
         _m.insert(StrPolyPair(name,vi->get_init()));
      }
   }

   

   return;
}

void IMap::add_initial_generators(GenVector & v){

   int i;

   for (i=0;i<_dim;++i){
      v.push_back(ray(Variable(i)));
      v.push_back(ray( -1 * Variable(i) ));

   }
   
}


bool IMap::entry_exists(TrLoc const & loc) const{
   return entry_exists(loc.get_name());
}

bool IMap::entry_exists(string const & name) const{
   return (_m.find(name)!=_m.end());
}


C_Polyhedron const & IMap::operator() (TrLoc const & loc) const{

   return operator()(loc.get_name());
   
}

C_Polyhedron const & IMap::operator() (string const & loc) const {

   PRECONDITION( entry_exists(loc), "Asked to retrieve a non-existent location"<<loc<<endl );

   return (_m.find(loc))->second;
}


C_Polyhedron & IMap::operator[] (TrLoc const & loc){

   return operator[](loc.get_name());
   
}

C_Polyhedron & IMap::operator[] (string const & loc){

   PRECONDITION( entry_exists(loc), "Asked to retrieve a non-existent location"<<loc<<endl );

   return (_m.find(loc))->second;
}

bool IMap::contains(IMap const & what) const{

 
   //
   // I am NOT checking for the compatibility
   // of the two maps
   //

   StrPolyMap::const_iterator vi;
   for(vi= _m.begin(); vi != _m.end(); ++vi){

      string const & name = vi -> first;
      C_Polyhedron const & p1 = vi -> second;
      C_Polyhedron const & p2 = what(name);

      if (!p1.contains(p2))
         return false;
   }

   return true;
}





void IMap::print(ostream & out) const{
   vector<TrLoc>::const_iterator vi;

   for (vi = _vloc.begin(); vi != _vloc.end(); ++vi){
      out << vi-> get_name() <<"[["<<endl;
      print_polyhedron(out,operator()(vi->get_name()), vi->get_dual_var_info());
      out<<"]]"<<endl;
   }
   
   return;
}


void IMap::print_using (ostream & out, var_info * f) const{
   vector<TrLoc>::const_iterator vi;

   for (vi = _vloc.begin(); vi != _vloc.end(); ++vi){
      out << vi-> get_name() <<"[["<<endl;
      print_polyhedron(out,operator()(vi->get_name()), f);
      out<<"]]"<<endl;
   }
   
   return;
}

ostream & operator<<(ostream & out, IMap const & what){
   what.print (out);
   return out;
}


GenVector & IMap::get_frames(string const & name){
   PRECONDITION( entry_exists(name), "Asked to retrieve a non-existent location"<<name<<endl );
   return (_frames.find(name))->second;
}

void IMap::widening_assign( IMap const & what) {
   vector<TrLoc>::const_iterator vi;

   for (vi=_vloc.begin(); vi!= _vloc.end(); ++vi){

      string const & name=vi->get_name();
      C_Polyhedron const & npoly= what(name);
      C_Polyhedron & mpoly = this->operator[](name);
      
      mpoly.H79_widening_assign(npoly);

   }
}

void IMap::propagation_assign (vector<TrRel> const & vrel, IMap & result){
   // for each location and each transition,
   // compute the post
   
   vector<TrRel>::const_iterator vi;
   
   for (vi=vrel.begin(); vi!= vrel.end(); ++vi){
      string const & preloc_name= (vi->get_preloc()).get_name();
      string const & postloc_name= (vi->get_postloc()).get_name();

      C_Polyhedron const & prepoly= this->operator()(preloc_name);
      C_Polyhedron temp(prepoly);

      C_Polyhedron  & postpoly= result[postloc_name];

      vi->post(prepoly,temp);

      postpoly.poly_hull_assign(temp);
   }
   //done?
}

void IMap::narrow_with(IMap & result) const{

   StrPolyMap::const_iterator vi;
   
   for(vi= _m.begin(); vi != _m.end(); ++vi){

      string const & name = vi -> first;
      C_Polyhedron const & p1 = vi -> second;
      C_Polyhedron & p2 = result[name];
      GenVector & res = result.get_frames(name);
      H79_narrow(p2,p1,res);
   }
      
   
}

IMap & IMap::operator=(IMap const & what){

   PRECONDITION( (_vloc==what._vloc), "Cannot assign with a mismatch in the location set");

   vector<TrLoc>::const_iterator vi;
   for (vi=_vloc.begin(); vi != _vloc.end(); ++vi){
      string const & name=vi->get_name();
      C_Polyhedron const & p1 = what(name);
      C_Polyhedron & p2 = operator[](name);

      p2 = p1;
      
   }

   return (*this);
}



void IMap::primalize(IMap & result) const{
   StrPolyMap::const_iterator vi;
   for(vi= _m.begin(); vi != _m.end(); ++vi){

      string const & name = vi -> first;
      C_Polyhedron const & p1 = vi -> second;
      C_Polyhedron & p2 = result[name];

      primal(p1,p2);
   }

   return;
}


void IMap::check(vector<TrRel> const & vrel) const{

   // first check for initiation

   vector<TrLoc>::const_iterator vi;
   bool failed=false;
   for (vi=_vloc.begin(); vi != _vloc.end(); ++vi){
      string const & name = vi->get_name();
      C_Polyhedron const & what = this->operator()(name);

      if (! what.contains(vi->get_init())){
         cerr<<" IMap::check --- Initiation failed for location "<<name<<endl;
         failed=true;
      }
         
   }

   // now check consecution

   vector<TrRel>::const_iterator vj;
   for (vj=vrel.begin(); vj!= vrel.end(); ++vj){
      
      string const & pre_name= (vj->get_preloc()).get_name();
      string const & post_name= (vj->get_postloc()).get_name();

      C_Polyhedron const & pre_poly = this->operator()(pre_name);
      C_Polyhedron const & post_poly = this->operator()(post_name);

      C_Polyhedron temp(pre_poly);
      temp.intersection_assign(vj->get_guard());

      C_Polyhedron result(post_poly);

      vj->compute_pre_poly(post_poly, result);

      if (!result.contains(temp)){
         cerr <<" Consecution failed for transition [[ "<<endl;
         cerr << (*vj)<<endl;
         cerr <<"]]"<<endl;
         // get a var_info
         var_info * f= vj->get_var_info();

         print_polyhedron(cerr,pre_poly,f);
         cerr<<endl;
         
         print_polyhedron(cerr,post_poly,f);
         
         
      }
   }
   

}
