
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#include "myassertions.h"
#include "Timer.h"
#include "LProblem.h"
#include <cmath>

#define _ASSERT_VALID_ROW(row) (    PRECONDITION ( (row >= 0 && row < _m) , " Row index "<< row<<" out of range") )
#define _ASSERT_VALID_COL(col) (    PRECONDITION ( (col >= 0 && col < _n) , " Col index "<< col<<" out of range") )




#define TOLERANCE_LIMIT ((double) 0.00001)

/* Are a and b within tolerance limits of each other */

#define WITHIN_LIMITS(a,b) ( ((double) a >= (double) b - TOLERANCE_LIMIT)  && \
                             ((double) a <= (double) b + TOLERANCE_LIMIT) )

#define ONE_TWEL(a) WITHIN_LIMITS( a, (double) 0.08333333333333333333)
#define TWO_TWEL(a) WITHIN_LIMITS( a, (double) 0.16666666666666666666)
#define THREE_TWEL(a) WITHIN_LIMITS( a, (double) 0.25)
#define FOUR_TWEL(a) WITHIN_LIMITS( a, (double) 0.333333333333333333333)
#define SIX_TWEL(a) WITHIN_LIMITS( a, (double) 0.5)
#define EIGHT_TWEL(a) WITHIN_LIMITS( a, (double) 0.66666666666666666666)
#define NINE_TWEL(a) WITHIN_LIMITS( a, (double) 0.75)
#define FIVE_SIX(a) WITHIN_LIMITS( a, (double) 0.833333333333333)

#define GET_DENOM(a) ( ONE_TWEL(a)? 1  :     \
		       TWO_TWEL(a)? 2  :     \
                       THREE_TWEL(a)? 3 :    \
                       FOUR_TWEL(a)? 4 :     \
                       SIX_TWEL(a)?  6:      \
                       EIGHT_TWEL(a)? 8:     \
                       NINE_TWEL(a)?  9: -1 )










void print_row(ostream & out, int index){

   out <<"r_"<<index<<" ";
}

#define START_TIMING(timer) Timer timer 
#define STOP_TIMING(timer) ( time_lin_programming += timer.compute_time_elapsed() )

int n_lin_program_calls;
int n_simplex_iterations;
long time_lin_programming;
int max_dimension_lp;
long tot_dimension_lp;

void print_col(ostream & out, int index){
   out <<"c_"<<index<<" ";
}

void print_bounds( ostream & out, map<int,double> const & tpr, string dirn, void (*printfn) ( ostream &, int)){

   map<int,double>::const_iterator vi;

   for (vi= tpr.begin(); vi != tpr.end(); ++vi){
      int i = vi->first;
      double v = vi -> second;

      printfn(out,i);
      out<< dirn << v <<endl;
   }
   
}


//
// the constructor
//

LProblem::LProblem(int m, int n): _m(m), _n(n), _maximize(false), _responsible(true) {

   // create a _lp
   
   _lp= lpx_create_prob();
   lpx_set_int_parm(_lp,LPX_K_MSGLEV,0);
   // now add all the rows and columns

   lpx_add_rows( _lp, _m);
   lpx_add_cols( _lp, _n);

   
}

//
// The copy constructor... careful here
//

LProblem::LProblem(LProblem const & c):_m(c._m),
                                       _n(c._n),
                                       _cm(c._cm),
                                       _colobjs(c._colobjs),
                                       _rowobjs(c._rowobjs),
                                       _upper_bounds_rows(c._upper_bounds_rows),
                                       _lower_bounds_rows(c._lower_bounds_rows),
                                       _upper_bounds_cols(c._upper_bounds_cols),
                                       _lower_bounds_cols(c._lower_bounds_cols),
                                       _responsible(false){
   
   
   // now build everything
   _lp=c._lp;
   
}


void LProblem::reset(int m , int n){

   INVARIANT ( _responsible, " do not reset copy constructed linear programs ");
   
   START_TIMING(timer);
   
   lpx_delete_prob(_lp);

   _lp= lpx_create_prob();
   lpx_set_int_parm(_lp,LPX_K_MSGLEV,0);
   // now add all the rows and columns
   
   // now add all the rows and columns

   lpx_add_rows( _lp, m);
   lpx_add_cols( _lp, n);

   
   _m=m;
   _n=n;
  

   
   INVARIANT ( lpx_get_num_rows(_lp)==_m, " Invariant failed "<< lpx_get_num_rows(_lp)<<" != "<<_m);
   INVARIANT ( lpx_get_num_cols(_lp)==_n, " Invariant failed");
   
   _cm.clear();
   _colobjs.clear();
   _rowobjs.clear();
   _upper_bounds_rows.clear();
   _upper_bounds_cols.clear();

   _lower_bounds_rows.clear();
   _lower_bounds_cols.clear();
  
   STOP_TIMING(timer);
}

void LProblem::make_all_rows_positive(){

   //
   // for each row, we set its lower bound to zero
   //
   
   int i;

   for(i=0;i<_m;++i)
      _lower_bounds_rows[i] = (double) 0;
   
}

void LProblem::make_row_positive(int row){

   PRECONDITION ( (row >= 0 && row < _m) , " Row index "<< row<<" out of range");

   _lower_bounds_rows[row]=(double) 0;
   
}


void LProblem::make_all_cols_positive(){

   //
   // for each col, we set its lower bound to zero
   //
   
   int i;

   for(i=0;i<_n;++i)
      _lower_bounds_cols[i] = (double) 0;
   
}

void LProblem::make_col_positive(int col){

   PRECONDITION ( (col >= 0 && col < _n) , " Col index "<< col<<" out of range");

  
   
   _lower_bounds_cols[col]=0.0;
   
}

void LProblem::set_row_lower_bound(int row, long what){

   _ASSERT_VALID_ROW(row);
   
   _lower_bounds_rows[row]= (double) what;
   
}

void LProblem::set_row_upper_bound(int row, long what){

   _ASSERT_VALID_ROW(row);

   _upper_bounds_rows[row]= (double)what;
   
}


void LProblem::set_col_lower_bound(int col, long what){

   _ASSERT_VALID_COL(col);

   _lower_bounds_cols[col]= (double)what;
   
}

void LProblem::set_col_upper_bound(int col, long what){

   _ASSERT_VALID_COL(col);

   _upper_bounds_cols[col]= (double)what;
   
}


void LProblem::set_coefficient(int row, int col, long what){
   
   _ASSERT_VALID_ROW(row);
   _ASSERT_VALID_COL(col);

   if (what==0)
      return;
   
   Coordinate c(row,col);

   if (get_coefficient(row,col)== (double) 0) {
      _cm.insert(                                             \
                  pair<Coordinate,double>(c, (double) what));
   } else {
      CoeffMap::iterator vi = _cm.find(c);
      vi->second= (double) what;
   }

   return;
}


void LProblem::add_to_coefficient(int row, int col, long  what){

   _ASSERT_VALID_ROW(row);
   _ASSERT_VALID_COL(col);

   if (what ==0 )
      return;
   
   Coordinate c(row,col);
   CoeffMap::iterator vi = _cm.find(c);
   if (vi == _cm.end() ) {
      _cm.insert( pair<Coordinate,double>(c, (double) what));
   } else {
      vi->second+= (double) what;
   }

   return;
}   


double LProblem::get_coefficient(int row, int col){

   
   _ASSERT_VALID_ROW(row);
   _ASSERT_VALID_COL(col);
   
   
   Coordinate c(row,col);
   
   CoeffMap::iterator vi = _cm.find(c);

   if (vi == _cm.end())
      return (double) 0;
   
   return vi->second;
   
}



void LProblem::set_objective_row(int row, long what){
      _ASSERT_VALID_ROW(row);
      PRECONDITION(what ==0, "GLPK no longer supports row objectives. Sorry!");
      
      _rowobjs[row]=(double) what;

      
}


void LProblem::set_objective_col(int col, long what){
      _ASSERT_VALID_COL(col);

      _colobjs[col]=(double) what;
}

void LProblem::prepare_objective_function(){
   // get rid of row objective coeffients.
   // For example if _rowobjs[i]=r_i then
   // _colobjs <- _colobjs + r_i * ( rowi  )

   
}

void LProblem::add_bounds (int i,
                           map<int,double> const & lower,
                           map<int,double> const & upper ,
                           void (* handlerfun) (LPX * , int, int, double, double)
                           ){

   map<int,double>::const_iterator vu =  upper.find(i);
   map<int,double>::const_iterator vl =  lower.find(i);

   double ub = (vu == upper.end())? (double) 0 : vu->second;
   double lb = (vl == lower.end())? (double) 0 : vl->second;

   int typx = LPX_FR;

   if (vu != upper.end()) {

      if (vl != lower.end()){

         if (lb == ub)
            typx = LPX_FX;
         else
            typx = LPX_DB;

      } else { typx= LPX_UP;}
      
   } else {
      if (vl != lower.end())
         typx = LPX_LO;
      else
         typx = LPX_FR;
   }

   //
   // my index i is known to the library
   // as i+1
   //
   
   handlerfun( _lp, i+1, typx, lb,ub);
   return;
}


void LProblem::add_coefficients_to_library(){

   int sz= _cm.size();

   int * rn = new int[1+sz];
   int * cl = new int[1+sz];
   double * a = new double[1+sz];
  
   CoeffMap::iterator vi;
   int i;
   
   for (i=1,vi = _cm.begin(); vi != _cm.end(); ++ vi, ++i ) {

      Coordinate const & c = vi->first;

      rn[i]=c.first+1;
      cl[i]=c.second+1;

      a[i]= vi->second;
   }

   lpx_load_matrix(_lp, sz, rn,cl, a);

   // This is speculatively saving memory
   // Does the library reallocate or does it use
   // my structures? Documentation is silent..
   // ayyyy mi querida core dump!.. tu no sabes cuanto yo te encanto!
   //
   
   delete(rn);
   delete(cl);
   delete(a);
   
   return;
}

void LProblem::add_objective_coefficients( int i, map<int, double> const & m,
                                           void (* setfunction) ( LPX * , int, double) ){

   map<int,double>::const_iterator vi = m.find(i);
   if (vi == m.end())
      return;

   setfunction(_lp, 1+i, vi->second);
   return;
}

void LProblem::load_problem(){
   START_TIMING(timer);
   //
   // load the problem into the library
   //

   //
   // First scan the rows and set the row constraints
   //
   
   int i;

   rebuild_bounds();
   
   //
   // now set up the coefficients
   // 

   add_coefficients_to_library();   

   //
   // now set up the objective direction
   //

   if (_maximize){
      lpx_set_obj_dir(_lp,LPX_MAX);
   } else
      lpx_set_obj_dir(_lp,LPX_MIN);

   //
   // now set up the objective function
   //

   /*
    * Code disabled. GLPK 4.7 does not support the
    * setting of objective row
    *
    */
   
   // for (i=0;i<_m;++i){
   //       add_objective_coefficients(i, _rowobjs, lpx_set_row_coef);
   //    }
   
   for (i=0;i<_n;++i){
      add_objective_coefficients(i, _colobjs, lpx_set_obj_coef);
   }

   STOP_TIMING(timer);
   
}


bool LProblem::is_feasible(){
   int status=lpx_get_status(_lp);
   
   // now interpret the status of the result
   bool aok=false;
   
   switch(status){
      case LPX_OPT:
         
      case LPX_FEAS:
                
      case LPX_UNBND:
         aok=true;
         break;

      case LPX_INFEAS:
         
      case LPX_NOFEAS:
           
      case LPX_UNDEF:

         aok=false;
         break;

   };

   return aok;
   

}

bool LProblem::is_unbounded(){
   int status=lpx_get_status(_lp);
   return (status == LPX_UNBND);
   
}


bool LProblem::interpret_status(){
   int status=lpx_get_status(_lp);
   
   // now interpret the status of the result
   bool aok=false;
   
   switch(status){
      case LPX_OPT:
         aok=true;
         break;
      case LPX_FEAS:
    
      
         break;

      case LPX_INFEAS:
    

         break;

      case LPX_NOFEAS:
    
         break;

      case LPX_UNBND:
    
         break;

      case LPX_UNDEF:
         cout<<" Undefined solution status "<<endl;
         break;

   };

   return aok;
   
}

bool LProblem::solve(){
   START_TIMING(timer);
   // simplex
    lpx_set_int_parm(_lp,LPX_K_MSGLEV,0);

    // add one to the # of calls
    n_lin_program_calls++;

    tot_dimension_lp+= _m+_n;
    // Check for the maximum dimensions
    if ( _m+_n > max_dimension_lp)
       max_dimension_lp=_m+_n;
    
    int n_iters = lpx_get_int_parm(_lp,LPX_K_ITCNT);
    int s1=lpx_simplex(_lp);
   n_iters = lpx_get_int_parm(_lp,LPX_K_ITCNT)- n_iters;
   
   n_simplex_iterations += n_iters; 
   
   if (s1 != LPX_E_OK){
      cerr<<" Fault flagged solving LP using simplex"<<endl;
      return false;
   }
   
   STOP_TIMING(timer);
   return true;
}

bool LProblem::solve_and_interpret_status(){
   START_TIMING(timer);
   solve();
   return interpret_status();
   STOP_TIMING(timer);
}

LProblem::~LProblem(){
   if (_responsible)
      lpx_delete_prob(_lp);
}

double LProblem::get_optimal_solution(){
   
  PRECONDITION( interpret_status(), " Problem does not seem to have optimal solution ");
 
  return lpx_get_obj_val(_lp);
}


long LProblem::get_optimal_value_truncated(){
   double Z=get_optimal_solution();
 
   if ( _maximize){
      // make sure that solution is a lower bound on the maximization
      return (long) floor(Z);
      
   } else {
      // make sure that the solution is an upper bound
     return (long) ceil(Z);
   }

   return (long)Z;
}

pair<long,long> LProblem::get_optimal_value_fractional(){
  
  /*
   * The algorithm is as follows:
   *       1. split the optimal value into the integer and fractional parts
   *       2. check if the fractional part is of a specific form within some 
   *          tolerance limits (arbitrarily set to (+,-) 10^-5
   *       3. If not, truncate the scaled optimal.
   *       4. Reduce results to the lowest terms
   */

  long num=0,den=1;

  double Z= get_optimal_solution();
    
  // split the optimal value into integer and mantissa
  
  double intpart;
  double mantissa;

  mantissa=modf(Z,&intpart);
 
  // now check if the mantissa is close to zero 
  
  if (WITHIN_LIMITS(mantissa, 0)){
    // the integer part is the result
    num= (long) intpart;
    den= 1;
    return pair<long,long>(num,den);
  }
  
  // make mantissa positive
  if (mantissa < 0) {
    mantissa = mantissa +1;
    intpart = intpart -1;
  }
  
  INVARIANT( ( mantissa > 0 && mantissa < 1) ,  \
	     "major invariant failure");
  
  
  if (WITHIN_LIMITS(mantissa, 1)){
    // the integer part is the result
    num= (long) intpart+1;
    den= 1;
  } else if (WITHIN_LIMITS(mantissa, 0)){
    // the integer part is the result
    num= (long) intpart;
    den= 1;
  } else if (WITHIN_LIMITS(mantissa, 0.33333333333)){
    num =  3 * (long) intpart + 1; 
    den = 3;
  } else if (WITHIN_LIMITS(mantissa, 0.5)){
    num =  2 * (long) intpart + 1; 
    den = 2;
  } else if (WITHIN_LIMITS(mantissa, 0.6666666666)){
    num =  3 * (long) intpart + 2; 
    den = 3;
  } else if (WITHIN_LIMITS(mantissa, 0.25)){
    num =  4 * (long) intpart + 1; 
    den = 4;
  } else if (WITHIN_LIMITS(mantissa, 0.75)){
    num =  4 * (long) intpart + 3; 
    den = 4;
  } else if (WITHIN_LIMITS(mantissa, 0.1)){
    num =  10 * (long) intpart + 1; 
    den = 10;
  } else if (WITHIN_LIMITS(mantissa, 0.2)){
    num =  5 * (long) intpart + 1; 
    den = 5;
  } else if (WITHIN_LIMITS(mantissa, 0.4)){
    num =  5 * (long) intpart + 2; 
    den = 5;
  } else if (WITHIN_LIMITS(mantissa, 0.6)){
    num =  5 * (long) intpart + 3; 
    den = 5;
  }else if (WITHIN_LIMITS(mantissa, 0.8)){
    num =  5 * (long) intpart + 4; 
    den = 5;
  } else if (WITHIN_LIMITS(mantissa, 0.83333333333)){
    num =  6* (long) intpart + 5; 
    den = 6;
  }else {

  
    // or else we just treat it differently for maximization and minimizations
#ifndef NDEBUG
    cerr<< " Message --- Encountered unrecognized mantissa : " << mantissa <<endl;
#endif
    if (_maximize){
    // we need to lower-bound the solution
      num= (long) intpart;
      den=1;
    } else{

      // or else we upperbound the solution
      num= (long) intpart + 1;
      den=1;
    }
  }
  INVARIANT( den > 0, " negative denominator can mess the whole process up ");
  
  return pair<long,long>(num,den);
}

void LProblem::rebuild_bounds_cols(){
      
  int i;

   for (i=0; i < _n ; ++i ){

      add_bounds(i, _lower_bounds_cols, _upper_bounds_cols, lpx_set_col_bnds);

   }
}

void LProblem::rebuild_bounds_rows(){
   int i;
   for (i=0; i < _m ; ++i ){

      add_bounds(i, _lower_bounds_rows, _upper_bounds_rows, lpx_set_row_bnds);

   }

}

void LProblem::rebuild_bounds(){
   rebuild_bounds_rows();
   rebuild_bounds_cols();
}


void LProblem::print_objective(ostream & out) const{
   map<int,double>::const_iterator vi;

   for (vi=_colobjs.begin(); vi!=_colobjs.end(); ++vi){

      int index=vi->first;
      double val = vi->second;

      out << val <<" * " ;
      print_col(out, index);
      out <<" + ";
   }
   
   for (vi=_rowobjs.begin(); vi!=_rowobjs.end(); ++vi){
      int index=vi->first;
      double val = vi->second;
      
      out << val <<" * " ;
      print_row(out, index);
      out <<" + ";
   }

   out << " 0 ";
}

void LProblem::print_problem(ostream & out ) const{
   //
   // get the expression for each row
   //

   int * cols = new int[_n+1];
   double * vals = new double[_n+1];

   for (int i=0; i<_m;++i){
      print_row(out,i);
      out <<" = ";
      int len=lpx_get_mat_row( _lp, 1+i, cols, vals);
      
      for (int j=1;j<= len ; ++j){

         int c = cols[j]-1;
         double v = vals[j];

         out <<v << " * ";
         print_col(out,c);
         out <<" + ";
         
      }
      out <<" 0 "<<endl;

      
   }
   delete(vals);
   delete(cols);
}

void LProblem::print( ostream & out) const {

   /*
    * Print the problem -- first print what needs to be maximized
    * 
    */
   if (_maximize)
      out<<" Max ";
   else
      out<<" Min ";
   
   print_objective(out);
   
   out <<endl;

   /*
    * Print the main problem
    */
   
   print_problem(out);

   /*
    * print bounds
    */

   print_bounds(out, _upper_bounds_rows, string("<="), print_row);
   print_bounds(out, _upper_bounds_cols, string("<="), print_col);
   print_bounds(out, _lower_bounds_rows, string(">="), print_row);
   print_bounds(out, _lower_bounds_cols, string(">="), print_col);

   // done
}

ostream & operator<< (ostream & out, LProblem const & lprob){
   lprob.print(out);
   return out;
}
