
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#ifndef D__LPROBLEM__H_
#define D__LPROBLEM__H_

#include <iostream>
#include <vector>
#include <string>
#include <map>
extern "C"{
#include <glpk.h>
}

using namespace std;

typedef pair<int,int> Coordinate;
typedef map<Coordinate, double> CoeffMap;
typedef pair<int,double> IntDoublePair;



/*
 *  Basic interface routine to the LP library.
 *  I am only interested in feasibility on the long run,
 *  but for the time being, I shall add provisions to
 *  solve a full blown LP problem
 */

class LProblem{
  private:
   /*
    * _m is the number of row variables,
    * _n is the number of column variables.
    *
    *  Important: Row indexing is from [0.._m-1]
    *             Col indexing is from [0.._n-1]
    *
    * _cm maps coordinates to non-zero double values
    * _maximize is true if the objective is to maximize
    * _rowobjs, _colobjs set the row and the column coefficients of the objective to the
    *  LP problem.
    *
    * _lp the problem itself
    *
    */

   

   int _m, _n;

   bool _maximize;
    
   CoeffMap _cm;   
   map<int, double> _colobjs, _rowobjs;   

   
   /*   lack of an upper bound makes the upper bound +\infty
    *   similarly lack of a lower bound makes it - \infty
    */
   
   map<int, double> _upper_bounds_rows;
   map<int, double> _lower_bounds_rows;
   map<int, double> _upper_bounds_cols;
   map<int, double> _lower_bounds_cols;

   LPX * _lp;

   bool _responsible;
   
   /*
    * Private functions
    *
    */
   void manufacture_row_name(string & x);
   void manufacture_col_name(string & x);
   
   void add_objective_coefficients( int i,
                                    map<int, double> const & m,
                                    void (* setfunction) ( LPX * , int, double) );
   void add_bounds (int i,
                    map<int,double> const & lower,
                    map<int,double> const & upper ,
                    void (* handlerfun) (LPX *, int, int, double, double));
   
   void add_coefficients_to_library();

   void print_problem(ostream & out)const ;

   void print_objective(ostream & out)const ;

   void prepare_objective_function();
   
  public:

   // start a problem
   
   LProblem(int m, int n);

   LProblem(LProblem const & lpr);
   
   // void add_extra_row();
   // void add_extre_column();
   // void remove_column();
   
   // add variable constraints

   void make_all_rows_positive();
   void make_all_cols_positive();
   
   void make_row_positive(int row);
   void make_col_positive(int col);
   
   void set_row_upper_bound(int row, long what);
   void set_row_lower_bound(int row, long what);
   
   void set_col_upper_bound(int col, long what);
   void set_col_lower_bound(int col, long what);

   // type of objective
   
   bool is_maximization() const{
      return _maximize;
   }

   void set_maximize(bool max=true){
      _maximize=max;
   }

   double get_coefficient(int row, int col);
   
   // set coefficients

   void set_coefficient(int row, int col, long  what);
   void add_to_coefficient(int row, int col, long  what);

   // set objective coefficients

   void set_objective_row( int row, long what);
   void set_objective_col(int col, long what);

   // load the problem into the library

   void load_problem();

   // solve the problem

   bool interpret_status();
   bool solve();
   bool is_feasible();
   bool is_unbounded();
   
   bool solve_and_interpret_status();

   // solve without rebuilding
   double get_optimal_solution();

   long get_optimal_value_truncated();
   pair<long,long> get_optimal_value_fractional();
   
   
   //bool check_solution_feasible();

   // print the problem
   void print (ostream & out) const;
     
   ~LProblem();

   // reset the problem

   void reset( int m, int n);

   // change bounds and rebuild bounds

   void  rebuild_bounds_cols();
   void rebuild_bounds_rows();
   void rebuild_bounds();

 
   
   
};

ostream & operator<< (ostream & out, LProblem const & lprob);


#endif
