
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include "LocTemplateMap.h"

int current_iteration;
int n_enrichments;
extern int num_equations_before_enrich;


void LocTemplateMap::initialize(TrLoc const & lc, EquationTemplate & what){

   C_Polyhedron const & init= lc.get_init();
   
   if (!lc.is_init_empty()){
      
      buildEquationTemplate(init,what);
      LProblem lpr(1,1);
      buildLProblem(init,lpr);
      
      // now consider a host of relations and optimize
      add_extra_relations(what,lpr);
      
   } else{
      
      what.set_inconsistent();
      add_extra_relations(what);
   }

}




void LocTemplateMap::populate_map(vector<TemplateBuilder>  &vtb){
   
   vector<TrLoc>::const_iterator vi;
   vector<TemplateBuilder>::iterator vj;
   
   for (vi=_vloc.begin(); vi != _vloc.end(); ++vi){

      string const & name=vi->get_name();
      
      EquationTemplate what(_n);
      
      
      // obtain the inital condition
      
      C_Polyhedron const & init= vi->get_init();

      // add the initial conditions
      if (!vi->is_init_empty()){
         
         buildEquationTemplate(init, what);
         // build a problem

         LProblem  lpr(1,1);
         buildLProblem(init, lpr);
         
         for (vj=vtb.begin(); vj!= vtb.end(); ++vj){
            vj->build_template(what,(lpr));
         }
         
      } else{
         what.set_inconsistent();

         for (vj=vtb.begin(); vj!= vtb.end(); ++vj){
            vj->build_template(what);
         }
      }
      
      _m.insert(StrTempPair(name,what));      
   }
}

void LocTemplateMap::populate_map() {
   
   vector<TrLoc>::const_iterator vi;

   for (vi=_vloc.begin(); vi != _vloc.end(); ++vi){

      string const & name=vi->get_name();
      
      EquationTemplate what(_n);
      
      initialize((*vi),what);

      _m.insert(StrTempPair(name,what));      
   }

   
   return;
}

void LocTemplateMap::populate_map(EquationTemplate & what){
   vector<TrLoc>::const_iterator vi;
   
   for (vi=_vloc.begin(); vi != _vloc.end(); ++vi){
      
      string const & name=vi->get_name();
    
      _m.insert(StrTempPair(name,what));      
   }
   
   
   return;
}


void LocTemplateMap::enrich_map_once(){

   /*
    * From existing relations, deduce "supporting" 
    * linear relationships.
    *
    */

   vector<TrRel>::const_iterator vi;
   // Iterate through transitions
   for (vi=_vrel.begin(); vi != _vrel.end(); ++vi){
      if (vi->is_petri_net()) continue;
      
      // Create a dummy lproblem
      
      LProblem lp(1,1);
      TrLoc const & pre = vi->get_preloc();
      TrLoc const & post = vi->get_postloc();

      string const & prename = pre.get_name();
      string const & postname = post.get_name();

      INVARIANT( _m.find(prename)!= _m.end(), " Location name does not have a map entry "<< prename );
      INVARIANT( _m.find(postname)!= _m.end(), " Location name does not have a map entry "<< postname );
      
      // get the pre and post equation templates
      EquationTemplate & pret = _m.find(prename) -> second;
      EquationTemplate & postt = _m.find(postname) -> second;

      // it pre template is inconsistent, or the pre's initial condition is empty
      if (pret.inconsistent() || (pre.get_init()).is_empty()){
         if (prename!=postname)
            enrich_pair(pret, postt, (*vi));
         else
            enrich_pair(pret, (*vi));
      } else {

         // or else decide if the transition is self-looping
         // or not and call the appropriate routing with the the LProblem
         // built from the initial condition.
         
         buildLProblem(pre.get_init(),lp);
         
         if (prename!=postname)
            enrich_pair(pret, postt, (*vi),lp);
         else
            enrich_pair(pret, (*vi), lp);
      }
   }
   
}


LocTemplateMap::LocTemplateMap(int n, var_info * f,
                               vector<TrLoc> const & vloc,
                               vector<TrRel> const & vrel):_n(n),
                                                           _f(f),
                                                           _vloc(vloc),
                                                           _vrel(vrel)
{
   populate_map();

   num_equations_before_enrich= get_total_template_size();
   
   for (int i=0;i< n_enrichments;++i){
      enrich_map_once();
   }
}

LocTemplateMap::LocTemplateMap(int n, var_info * f,
                               vector<TrLoc> const & vloc,
                               vector<TrRel> const & vrel,
                               vector<TemplateBuilder>  & vtb):_n(n),
                                                                    _f(f),
                                                                    _vloc(vloc),
                                                                    _vrel(vrel)
{

   // build map
   
   if (vtb.empty())
      populate_map();
   else
      populate_map(vtb);

   num_equations_before_enrich= get_total_template_size();
   
   // enrich map
   for (int i=0;i< n_enrichments;++i){
      enrich_map_once();
   }
}



LocTemplateMap::LocTemplateMap(int n, var_info *f, \
                               vector<TrLoc> const & vloc, \
                               vector<TrRel> const & vrel, \
                               EquationTemplate  & what):_n(n),
                                                              _f(f),
                                                              _vloc(vloc),
                                                              _vrel(vrel){

   // map each location to the template "what"
   // secretly we know there is going to be just one location here

   PRECONDITION( _vloc.size()==1, " This has only been implemented for one location systems -- sorry");

   populate_map(what);

   num_equations_before_enrich= get_total_template_size();
   
   //done!!
}
   
bool LocTemplateMap::post(TrRel const & what){
   
   StrTempMap::iterator vi = _m.find((what.get_preloc()).get_name());
   StrTempMap::iterator vj = _m.find((what.get_postloc()).get_name());
   
   PRECONDITION ( ( vi != _m.end() && vj != _m.end()), " Illegal pre/post location here");

   EquationTemplate & pre = vi->second;
   EquationTemplate & pos= vj->second;

   /*
    * A post only needs to be computed if the
    * pre location was updated in the current or the
    * previous iterations
    *
    */
   
   if (pre.iteration_count() >= current_iteration -1 ){
      post_computation(pre, pos, what);
   }

   return true;
}


void LocTemplateMap::compute_invariants(){

   current_iteration=0;
   
   vector<TrRel>::const_iterator vi;
   bool something= true;
   StrTempMap::iterator vj;
   
   while(something){
      current_iteration++;
      
      for (vi=_vrel.begin(); vi != _vrel.end() ; ++vi){
         post((*vi));
      }
      something=false;
      
      cerr <<" The map is :"<<endl;
      cerr << (*this)<<endl;

      
      
      
      for (vj=_m.begin(); vj != _m.end(); ++vj){
         EquationTemplate & w= vj->second;
         
         if (w.iteration_count()==current_iteration){   
            /*
             * At least One more step is required
             */

            something=true;
         }
      }
      
   }
}


void LocTemplateMap::print(ostream & out) const{
   StrTempMap::const_iterator vi;
 
   out << " Iteration # " << current_iteration<<endl;
  
   for (vi = _m.begin(); vi != _m.end(); ++vi){
      
      out << (vi -> first) << "[["<<endl;
      (vi->second).print_using( out, _f );
      out<<"]]"<<endl;
      
   }
   out << "================================================="<<endl;
   
   return;
   

}

ostream & operator<< (ostream & out, LocTemplateMap const & ltm){
   ltm.print(out);
   return out;
}



void LocTemplateMap::fill_imap( IMap & result ) const{

   StrTempMap::const_iterator vi;
   
   for (vi = _m.begin(); vi != _m.end(); ++vi){
      string const & name= vi->first;

      C_Polyhedron & what = result[name];
      
      what = C_Polyhedron(_n, UNIVERSE);

      (vi->second).convert(what);
      
   }

   return;
   
}

int LocTemplateMap::get_total_template_size() const{
   StrTempMap::const_iterator vi;
   int total=0;
   for (vi=_m.begin(); vi != _m.end(); ++vi){
      total+=(vi->second).get_num_equations();

   }
   return total;
}
