
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include <iostream>
#include <map>
#include "funcs.h"
#include "myassertions.h"
#include "PolyUtils.h"
#include "SparseLinExpr.h"

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;


void SparseLinExpr::clear_out(){
   m_.clear(); // that should erase all the entries!
   
}

void SparseLinExpr::initialize(int n, var_info * f){
   n_ =n;
   f_ =f;
   m_.clear();
   count_ =1;
   info_set_=true;
}


bool SparseLinExpr::_class_invariant_check() const{
   IRMap::const_iterator vi;
   if (m_.empty()) return true;
   for (vi=m_.begin(); vi !=m_.end(); ++vi){
      if ((*vi).first <0 || (*vi).first > n_){
         cerr<<" Illegal Dimensions!"<<endl;
         return false;
      }
      if ((*vi).second ==0){
         cerr<<"Zero Entry"<<endl;
         return false;
      }
   }

   return true;
}


void SparseLinExpr::add_coefficient(int index, Rational const & what){

   //
   // this will insert the index if it has not been introduced by
   // default The constructor Rational() will be called-- that
   // produces a Rational that is zero by default and then G.W. things
   // should work fine!
   //
   
   
   
   PRECONDITION( (index >= 0 && index < n_+1 ),                      \
                 "SparseLinExpr::add_coefficient --> illegal index");

   if (what==0) return;

   
   Rational & r = m_[index];

   r+=what;
   
   if (r==0){
      // remove the position
      m_.erase(index);
   }
   
   return;   
}


void SparseLinExpr::subtract_coefficient(int index, Rational const & what){

   //
   // this will insert the index if it has not been introduced by
   // default The constructor Rational() will be called-- that
   // produces a Rational that is zero by default and then G.W. things
   // should work fine!
   //
   
   
   
   PRECONDITION( (index >= 0 && index < n_+1 ),                      \
                 "SparseLinExpr::subtract_coefficient --> illegal index");

   if (what==0) return;

   
   Rational & r = m_[index];

   r-=what;
   
   if (r==0){
      // remove the position
      m_.erase(index);
   }
   
   return;   
}


void SparseLinExpr::subtract_scaled(SparseLinExpr const & x, Rational const & scale){
   IRMap const & xm= x.get_map();
   IRMap::const_iterator vi;
   for(vi=xm.begin(); vi !=xm.end(); ++vi){
      subtract_coefficient((*vi).first,(*vi).second * scale);
   }
   return;
}


void SparseLinExpr::add_scaled(SparseLinExpr const & x, Rational const & scale){
   IRMap const & xm= x.get_map();
   IRMap::const_iterator vi;
   for(vi=xm.begin(); vi !=xm.end(); ++vi){
      add_coefficient((*vi).first,(*vi).second * scale);
   }
   return;
}

void SparseLinExpr::set_coefficient(int index, Rational const & what){

   //
   // this will insert the index if it has not been introduced by
   // default The constructor Rational() will be called-- that
   // produces a Rational that is zero by default and then G.W. things
   // should work fine!
   //
   
   
   
   PRECONDITION( (index >= 0 && index < n_+1 ),                      \
                 "SparseLinExpr::set_coefficient --> illegal index");


   
   Rational & r = m_[index];

   r=what;
   
   if (r==0){
      // remove the position
      m_.erase(index);
   }
   
   return;   
}





//
// This is inline only for this particular file
//

inline Rational  SparseLinExpr::get_coefficient(int index) const{

   PRECONDITION( (index >= 0 && index < n_+1 ),                      \
                 "SparseLinExpr::get_coefficient --> illegal index"<<index);

   IRMap::const_iterator pos= m_.find(index);

   if (pos == m_.end()) return Rational();

   return (*pos).second;
   
}

SparseLinExpr::SparseLinExpr():n_(0),
                               count_(1),
                               info_set_(false)
{}

SparseLinExpr::SparseLinExpr(int n, var_info * f): n_(n),
                                                   f_(f),
                                                   count_(1),
                                                   info_set_(true)
{
   PRECONDITION( n <= f_->get_dimension(), " Not enough print info ");
}

SparseLinExpr::SparseLinExpr(LinExpr const & ll): n_(ll.get_dim()),
                                                  f_(ll.get_info()),
                                                  count_(ll.get_count()),
                                                  info_set_(true){
   int i;

   for (i=0;i<n_+1;i++){
      if (ll(i)!=0) {
         add_coefficient(i,ll(i));
      }
      
   }
}




SparseLinExpr SparseLinExpr::operator+ (SparseLinExpr const & p1) const{

   //
   // I am adding this restriction ftb. If it proves to be a p.i.b then I will
   // remove it in later versions
   //

   PRECONDITION( (p1.get_dim()== n_) , " SparseLinExpr::operator+ : trying to add expressions of different size?" );
   
   SparseLinExpr ret_val(n_,f_);

   if (!is_printable())
      ret_val.make_unprintable();
   
   // for the time being, I will just iterate through all the elements in p1 and p2 and
   // add them to ret_val

   Rational tmp;
   
   IRMap::const_iterator vi;

   for (vi=m_.begin(); vi!=m_.end(); vi++){
      ret_val.add_coefficient( (*vi).first ,(*vi).second );
      
   }
   
   IRMap const & mp=p1.get_map();
   
   IRMap::const_iterator vj;

   for (vj=mp.begin(); vj!=mp.end(); vj++){
      ret_val.add_coefficient( (*vj).first ,(*vj).second);
   }

   //ret_val.prune();
   
   return ret_val;
}

SparseLinExpr SparseLinExpr::operator- (SparseLinExpr const & p1) const{

   //
   // I am adding this restriction ftb. If it proves to be a p.i.b then I will
   // remove it in later versions
   //

   PRECONDITION( (p1.get_dim()== n_) , " SparseLinExpr::operator+ : trying to add expressions of different size?" );

   SparseLinExpr ret_val(n_,f_);

   if (!is_printable())
      ret_val.make_unprintable();
   
   // for the time being, I will just iterate through all the elements in p1 and p2 and
   // add them to ret_val

   Rational tmp;
   
   IRMap::const_iterator vi;

   for (vi=m_.begin(); vi!=m_.end(); vi++){
      ret_val.add_coefficient( (*vi).first ,(*vi).second );
      
   }
   
   IRMap const & mp=p1.get_map();
   
   IRMap::const_iterator vj;

   for (vj=mp.begin(); vj!=mp.end(); vj++){
      ret_val.subtract_coefficient( (*vj).first ,(*vj).second);
   }

   //ret_val.prune();
   
   return ret_val;
}

SparseLinExpr & SparseLinExpr::operator+= (SparseLinExpr const & sl){

   PRECONDITION ((sl.get_dim() == n_) ,
                 "SparseLinExpr::operator+=  -- addition over incompatible dimensions");
   
   IRMap::const_iterator vj;
   IRMap const & mp=sl.get_map();

   for (vj=mp.begin(); vj !=mp.end();++vj){

      add_coefficient( (*vj).first, (*vj).second);

   }
   return (*this);
}

SparseLinExpr & SparseLinExpr::operator-= (SparseLinExpr const & sl){

   PRECONDITION ((sl.get_dim() == n_) ,
                 "SparseLinExpr::operator-= subtraction over incompatible dimensions");
   
   IRMap::const_iterator vj;
   IRMap const & mp=sl.get_map();

   for (vj=mp.begin(); vj !=mp.end();++vj){

      subtract_coefficient( (*vj).first, (*vj).second);

   }
   
   return (*this);
}



void SparseLinExpr::merge_assign(SparseLinExpr & ex2){
   int n2=ex2.get_dim(),n1=n_;
   //
   // I am not creating a new var_info
   // any attempt to print this may not go well
   //
   
   n_+=n2; // increase the dimensions!!

 
   IRMap const & m2=ex2.get_map();
   IRMap::const_iterator vi;
   for (vi=m2.begin(); vi!= m2.end();vi++){
      set_coefficient( (*vi).first + n1, (*vi).second);
   }
   
   return;
}


SparseLinExpr & SparseLinExpr::operator= (SparseLinExpr const & p1){

   initialize(p1.get_dim(),p1.get_info());
   
   count_=p1.get_count();

   IRMap::const_iterator vi;
   IRMap const & mp=p1.get_map();

   for(vi=mp.begin();vi!=mp.end();++vi){

      set_coefficient((*vi).first, (*vi).second);
   }
   
   return (*this);
}


SparseLinExpr& SparseLinExpr::operator*= (Rational const & r){
   
   if (r==0) {
      clear_out();
      return (*this);
   }
   
   IRMapIterator vi;

   for (vi=m_.begin(); vi!=m_.end(); ++vi){

      INVARIANT( ( (*vi).second != 0 ) ,\
                 "Class SparseLinExpr -- Invariant violated");
      
      ((*vi).second) *= r;
   }

   return (*this);
}


SparseLinExpr& SparseLinExpr::operator*= (int i){
   if (i==0){
      clear_out();
      return (*this);
   }
   IRMapIterator vi;

   for (vi=m_.begin(); vi!=m_.end(); ++vi){

      INVARIANT( ((*vi).second!=0) , "Class SparseLinExpr -- Invariant violated");
      
      ((*vi).second) *= i;
   }

   return (*this);
}

bool SparseLinExpr::operator== (SparseLinExpr const & t ) const{
   
   const IRMap & mp= t.get_map();

   return (m_ == mp); // That should do it?
}


/* 
 * Rational & SparseLinExpr::operator[] (int index){
 *
 *  //
 *  // This is dangerous..
 *  // Playing around with it may cause
 *  // the class to violate an invariant.
 *  //
 *  
 *  PRECONDITION( (0 <= index && index < n_+1), \
 *                "SparseLinExpr::operator[] --> reference attempt out of range"<<index );
 *  
 *  PRECONDITION ( (false), \
 *               " Using SparseLinExpr::operator[]. Please use set_coefficient/add_coefficient instead" );
 *   
 *  
 *  
 *  return m_[index];
 *  
 * }
 */


Rational SparseLinExpr::operator() (int index) const{
   return get_coefficient(index);
}

bool SparseLinExpr::is_constant() const{

   INVARIANT ( ( _class_invariant_check() ) , \
               "Class SparseLinExpr:: Invariant violation");
    
   if (m_.size() ==0) return true;
   if (m_.size() >1) return false;
   IRMap::const_iterator vt=m_.begin();
   
   // Is the only entry for the constant?
   if ( (*vt).first == n_ )
      return true;

   return false;
}



int SparseLinExpr::get_denominator_lcm() const{
   IRMap::const_iterator vi;
   IRPair pir;
   int run=1;
   for (vi=m_.begin();vi !=m_.end();++vi){
      pir=(*vi);
      INVARIANT ( (pir.second !=0 ), \
                  " Class SparseLinExpr --> Invariant violated. Key "<<pir.first<<" has zero entry " );

      run=lcm(run, (pir.second).den());
      
   }
   
   return run;
}


int SparseLinExpr::get_numerator_gcd() const{

   int run=1;
   bool first_number_seen=false;

   IRMap::const_iterator vi;
   IRPair pir;

   for (vi=m_.begin();vi !=m_.end();++vi){
      pir=(*vi);
      INVARIANT ( (pir.second !=0 ), \
                  " Class SparseLinExpr --> Invariant violated. Key "<<pir.first<<" has zero entry " );
      if (!first_number_seen){
         first_number_seen=true;
         run= (pir.second).num();
      } else 
         run=gcd(run, (pir.second).num());
      
   }
   
   return run;
}


bool SparseLinExpr::equiv( SparseLinExpr const & l1, Rational & factor) const{

   // check if there is a factor such that factor * this = l1

   if ( l1. is_zero()) {
      factor = 0;
      return true;
   }

   // If l1 is zero and I am not, return false
   
   if (is_zero())
      return false;
   
   // now check if for each entry in me, the corresponding entry in l1 behaves OK

   IRMap const & mp =  l1.get_map();
   IRMap::const_iterator vi, vj;
   // If sizes are not the same, then assuming the class invariant, there is
   // no way

   // as a lark, I will check the class invariant here

   INVARIANT ( ( _class_invariant_check() ) , \
               "Class SparseLinExpr:: Invariant violation");
   
   if (mp.size()!=m_.size()){
      factor=0;
      return false;
   }

   //now find a factor and check for it
   vi=m_.begin();
   vj=mp.begin();
   if (vi->first!= vj->first){
      return false;
   }
      
   factor = ((*vi).second).inv() * ( (*vj).second);

   for ( ; vi != m_.end() && vj != mp.end() ; ++vi, ++vj) {
      if  ( ( vi->first != vj-> first) || ( (factor * (*vi).second ) != (*vj).second ) ) 
         return false;
      
   }
   
   return true;
}





void SparseLinExpr::print(ostream & out) const{

   PRECONDITION(info_set_,
                " asked to print an Expression without a valid var-info ");

   
   INVARIANT ( ( _class_invariant_check() ) , \
               "Class SparseLinExpr:: Invariant violation");

   // print this stuff

   if (is_zero()){
      out <<" 0 ";
      return;
   }
   if (is_constant()){
      out<< get_coefficient(n_);
      return;
   }
   
   IRMap::const_iterator vi=m_.begin();
   
   out << (*vi).second << "*" << f_->get_name((*vi).first) ;
   vi++;
   for (;vi !=m_.end();++vi){

      INVARIANT( ((*vi).second != 0)        ,\
                 "Class SparseLinExpr ::print() --> Invariant violated");
      
      if (! ((*vi).second < 0) )
         out <<"+";

      if (vi->first < n_)
         out << (*vi).second << "*" << f_->get_name((*vi).first) ;
      else
         out << (*vi).second ;
   }

   

}


void SparseLinExpr::add_to_lin_expression(Linear_Expression & ll, int scale_fact, int offset) const {
   // nothing to be done if I am zero
   if (is_zero()) return;
   // first obtain a scale factor to convert each coefficient to an integer
   Rational factor(get_denominator_lcm(), get_numerator_gcd());

   // now iterate through the expression
   IRMap::const_iterator vi = m_.begin();
   int i,j;
   Rational r;
   for(;vi!=m_.end();++vi){
      i = (*vi).first;
      r= (*vi).second * factor;
      
      INVARIANT( r.is_integer()          ,\
                 " SparseLinEspr::add_to_lin_expression --> gcd and lcm functions in SparseLinExpr seem buggy " );

      // obtain the appropriate coefficient by adding the scale factor
      j=r.num() * scale_fact;
      ll+= j * Variable(i+offset);
   }
   // done
   return;

}


void SparseLinExpr::occurring_variables(set<int> & what, int k) const{
   
   IRMap::const_iterator vi;
   for (vi=m_.begin(); vi !=m_.end();++vi) {
      if ((*vi).first>=k)
         what.insert((*vi).first);
   }
   return;
}

void SparseLinExpr::add_variables_in_front(int n1){

   PRECONDITION ( (n1 >=0) , "Asked to remove variables in the guise of adding variables");
   
   make_unprintable(); // 'coz I do not have a valid var-info anymore
   n_+=n1;
   IRMap m1;
   IRMap::iterator vi;
   int i;
   Rational r;
   for( vi= m_.begin(); vi !=m_.end(); ++vi){
      i=(*vi).first;
      r=(*vi).second;
      m1[i+n1]=r; // add 
   }
   m_.erase(m_.begin(), m_.end());
   m_=m1;// do an assignment
   
}


void SparseLinExpr::add_variables_in_back(int n1){

   PRECONDITION ( (n1 >=0) , "Asked to remove variables in the guise of adding variables");

   if (f_->get_dimension() < n_+n1){
      make_unprintable();
   }
   
   n_+=n1;
   // that is it
   
}


Rational SparseLinExpr::evaluate(Generator const & g)const {
   IRMap::const_iterator vi;
   Rational rt(0,1),c;
   int i,j;
   for(vi=m_.begin(); vi != m_.end(); ++vi ){

      i=(*vi).first;
      
      c=(*vi).second;

      if (i < n_) {
         // If it is not the constant term, find its coefficient and
         // multiply with mine
         j=handle_integers(g.coefficient(Variable(i)));
         c*=Rational(j,1);
      }
      // add to the return value
      rt+=c;
      
   }

   return rt;
}




ostream & operator<< (ostream & os, SparseLinExpr const & sp){
   sp.print(os);
   return os;
}


SparseLinExpr operator* (Rational const & i, SparseLinExpr const & p1){
   SparseLinExpr p(p1);
   p*= i;
   return p;
}

SparseLinExpr operator* (SparseLinExpr const & p1, Rational const & i){
   SparseLinExpr p(p1);
   p*= i;
   return p;
}

SparseLinExpr operator* (int i, SparseLinExpr const & p1){
   SparseLinExpr p(p1);
   p*= i;
   return p;
}



Linear_Expression SparseLinExpr::to_lin_expression() const{
   int j=get_denominator_lcm();
   int num,den;

   num=get_coefficient(n_).num();
   den=get_coefficient(n_).den();
   
   Linear_Expression l( num*j/den ); // set the constant term
   for (int i=0;i<n_;i++){
      num=get_coefficient(i).num();
      den=get_coefficient(i).den();
      l= l+ (num*j/den) * Variable(i);
   }
   return l;
      
}


Constraint SparseLinExpr::get_constraint(int ineq_type) const{
   Linear_Expression l = to_lin_expression();
   switch(ineq_type){
      case TYPE_LEQ:
         return l <= 0;
      case TYPE_EQ:
         return l==0;
      case TYPE_GEQ:
         return l >= 0;
      case TYPE_GE:
         return l > 0;
      case TYPE_LE:
         return l < 0;
         
   }
   return l==0; // by default
}


SparseLinExpr operator* (SparseLinExpr const & p1, int i){
   SparseLinExpr p(p1);
   p*= i;
   return p;
}


