
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */



#include "myassertions.h"
#include "TemplateBuilder.h"


int TemplateBuilder::get_count() const{

   int ret=0;
   
   vector<CoeffPair>::const_iterator vi;

   for (vi=_v.begin(); vi!= _v.end(); ++vi){

      if (vi->second < 0 ||  vi->second >= _n)
         ret++;

   }
   
   return ret;
   
}

void TemplateBuilder::obtain_instances(vector<Eqn> & result){

   vector<CoeffPair> instances;

   obtain_instances_rec(_v.begin(), instances, result);
}

void TemplateBuilder::obtain_instances_rec( vector<CoeffPair>::const_iterator vw, \
                                            vector<CoeffPair> & instances, \
                                            vector<Eqn> & result){
   // check the base case
   if (vw == _v.end()){
      // create the Eqn
      Eqn res(_n);
      instantiate_eqn(instances, res);
      // add it to result
      result.push_back(res);
      
      return;
   }

   // now the recursion

   // if (vw) does not need instantiation

   if ( vw->second >=0 && vw ->second < _n) {
      
      // just push the instance and continue with the recursion
      instances.push_back((*vw));

      obtain_instances_rec(vw+1, instances, result);

      // remove the last instance.
      instances.pop_back();
      
   }  else {

      // iterate through all the possibilities for the first variable
      int i;

      for (i=0;i<_n;++i){

         instances.push_back(CoeffPair(vw->first,i));

         obtain_instances_rec(vw+1, instances, result);

         instances.pop_back();
      }

      return;
   }
   

   
}

void TemplateBuilder::instantiate_eqn(vector<CoeffPair> const & instances,             \
                                      Eqn & result){
   
   vector<CoeffPair>::const_iterator vi;

   for (vi= instances.begin(); vi!= instances.end(); ++vi){

      PRECONDITION(vi->second >=0 && vi-> second < _n , \
                   " Variable out of range in TemplateBuilder::instantiate_eqn(..)");

      result[vi->second]+=vi->first;
      
   }

   return;
}


void TemplateBuilder::build_template(EquationTemplate & result){

   PRECONDITION( result.inconsistent(), \
                 " Add to consistent store without any LProblem? How?");

   // obtain all the instances
   vector<Eqn> vecr;

   obtain_instances(vecr);

   // add them one by one
   
   vector<Eqn>::iterator vi;

   for (vi=vecr.begin(); vi!= vecr.end(); ++vi){

      if (vi->is_zero()) continue;

      result.add_equation((*vi));
   }
   
}

void TemplateBuilder::build_template(EquationTemplate & what, LProblem & lpr){
   
   PRECONDITION( !what.inconsistent(), \
                 " Add to inconsistent store with a LProblem? Why?");

   
   vector<Eqn> vecr;

   obtain_instances(vecr);
   
   vector<Eqn>::iterator vi;

   for (vi=vecr.begin(); vi!= vecr.end(); ++vi){

      if (vi->is_zero()) continue;

      optimize_and_add((*vi), what, lpr);

   }  
}


void TemplateBuilder::print(ostream & out) const{

   vector<CoeffPair>::const_iterator vi;

   for (vi= _v.begin(); vi != _v.end(); ++vi){
      out << "( " << vi->first <<" , " << vi->second <<" )"<<endl;
   }

   return;
}


ostream & operator<< ( ostream & out, TemplateBuilder const & what){
   what.print(out);
   return out;
}


