
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#ifndef D__TEMPLATE_BUILDER__H_
#define D__TEMPLATE_BUILDER__H_

/*
 * A template builder that stores a pattern and
 * proceeds to create an "EquationTemplate"
 * This is primarily meant to replace the
 * functions that build templates in the LocTemplateMap
 * slowly all the functionality provided by the functions
 *     LocTemplateMap::enrich_map_once 
 *     LocTemplateMap::initialize()
 *     LocTemplateMap::add_default_pairs () ( there will be no more default pairs)
 *  and their slave functions will be moved into this class.
 *
 */

#include <iostream>
#include <map>
#include <string>
#include "var-info.h"
#include "Rational.h"
#include "TrLoc.h"
#include "TrRel.h"
#include "TrSys.h"
#include "IMap.h"

#include "LProblem.h"
#include "Eqn.h"
#include "EquationTemplate.h"
#include "EqnUtilities.h"

typedef pair<int, int> CoeffPair;

class TemplateBuilder{
  private:
   /*
    *  _n : The dimensionality
    *  _v :  A vector of CoeffPair
    *      each coeff-pair (i,j) means
    *              i * variable(j) if j >=0 && j<_n
    *              i * place_holder otherwise
    */

   int _n;
   vector<CoeffPair> _v;

   /* get_count() is not required anymore.
    * just leaving it in.
    */
    int get_count() const;
   /*
    */
    
   void instantiate_eqn(vector<CoeffPair> const & instances, Eqn & result);

   void obtain_instances(vector<Eqn> & result);

   void obtain_instances_rec( vector<CoeffPair>::const_iterator vw,  \
                              vector<CoeffPair> & instances, \
                              vector<Eqn> & result);
   

   
   
  public:
   
   /*  A simple constructor
    */ 

   TemplateBuilder(int n): _n(n){
   }
   
   /* Build a set of Eqn's and return
    * optimized w.r.t a given linear program
   */

   void insert_pair(int c, int var){
      if (c!=0)
         _v.push_back(CoeffPair(c,var));
   }
   
   void build_template(EquationTemplate & result, LProblem & lpr);
   /*
    * Without the linear program if necessary
    * The template better have the inconsistent flag set if calling this
    */

   void build_template(EquationTemplate & result) ;
   
   /*
    * print the template
    */
   
   void print(ostream & out) const;

   
   
};

ostream & operator<< (ostream & out, TemplateBuilder const & what);

#endif
