
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#include "myassertions.h"
#include "funcs.h"
#include "TrRelImpl.h"

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;


TrRelImpl::TrRelImpl(TrLoc & prel, TrLoc & postl,
             C_Polyhedron const & g):
      _n(prel.get_dimension()),
      _f(prel.get_var_info()),
      _prel(prel),
      _postl(postl),
      _guard(g),
      _dguard(_n,UNIVERSE),
      _breached(false),
      _petri(true){
   
   // if (dualize_guard)
   // dualize(_guard,_dguard);
   initialize_dual_update();
   //_guard_gens=_dguard.minimized_generators();
      
      
}

         
// the copy constructor

TrRelImpl:: TrRelImpl(TrRelImpl const & what): _n(what._n),
                                   _f(what._f),
                                   _prel(what._prel),
                                   _postl(what._postl),
                                   _guard(what._guard),
                                   _dguard(what._dguard),
                                   _update(what._update),
                                   _dupdate(what._dupdate),
                                   _breached(what._breached),
                                   _petri(what._petri),
                                   _name(what._name)
{}

bool TrRelImpl::update_exists(int i) const{
   return (_update.find(i)!=_update.end());
}


void TrRelImpl::initialize_dual_update(){

   //
   // to start off.. add entries to _dupdate
   // set everything to zero
   //
   
   int j;
   for (j=0; j < _n;++j){
      _dupdate.insert(ILEPair(j,Linear_Expression(0)));
   }
   _dupdate.insert(ILEPair(_n, Linear_Expression(Variable(_n)) ));

   return;
}

void TrRelImpl::add_update(int i, Linear_Expression const & what) {

   //
   // check if an update relation corresponding to i already
   // exists
   //

   PRECONDITION( (! update_exists(i)),
                 "Precondition violation -- asked to add an update when one has already been defined");
   
   _update.insert(ILEPair(i,what));

   // now insert into the dual update
   
   for (int j=0;j<_n;++j){
      if (j!=i ){

         if (what.coefficient(Variable(j))!= 0 ){
            _petri=false;
         }

      } else {
         if (what.coefficient(Variable(j))!=1){
            _petri=false;
         }
      }
      Linear_Expression & dexpr =  ( *( _dupdate.find(j)) ).second;
      dexpr += what.coefficient(Variable(j)) * Variable(i);
   }

   
   Linear_Expression & cexpr =  ( *( _dupdate.find(_n)) ).second;
   cexpr+= what.inhomogeneous_term() * Variable(i);

   return;   
}

void TrRelImpl::add_constant_updates(){

   //
   // whatever variables do not have an update relation
   // fill them up with the constant update
   //

   int i;
   for (i=0;i<_n;++i)
      if (! update_exists(i))
         add_update(i, Linear_Expression(Variable(i)));

   return;
}

void TrRelImpl::print(ostream & out) const{

   //
   // pretty print the transition
   //

   // first print the guard

   out <<" TrRelImpl: "<<_prel.get_name()<< " --> " << _postl.get_name()<<endl;
   
   out<<" Guard [[ "<<endl;
   print_polyhedron(out,_guard,_f);
   out <<"]]"<<endl;

   out<< "Updates [["<<endl;
   for (int i=0;i<_n;++i){
      if (update_exists(i)){
         out<< _f->get_name(i)<<" := ";
         print_lin_expression(out, (_update.find(i))->second , _f);
         out<<endl;
      }
   }
   out <<"]]"<<endl;
   out << " Transition is  a Petri Net?";

   if (_petri)
      out<<" yes!"<<endl;
   else
      out<<" no!"<<endl;
   
   return;
   
}


ostream & operator<< (ostream & out, TrRelImpl const & what){
   what.print(out);
   what.print_dual(out);
   return out;
}




bool TrRelImpl::guard_breached(C_Polyhedron const & what){

   if (_breached) return true;

   //
   // check if dual(what) is incompatible with dual(guard)
   // by checking if (0....,-1) is a point inside what
   // i.e., check to see if there is a constraint of the from
   // .... + <pos> * d >=0
   // or
   // ... + <non-zero> * d == 0
   // if so, then (0,...,0,-1) is clearly not inside
   // if not then (0,....,0,-1) is clearly inside
   //
   

   Constraint_System const & cs=what.constraints();
  
   Constraint_System::const_iterator vi;
  
   for (vi=cs.begin(); vi != cs.end(); ++vi){
      if ((*vi).is_inequality() && sgn((*vi).coefficient(Variable(_n))) > 0 ){
         _breached=true;
         return true;
      }else if ((*vi).is_equality() && sgn((*vi).coefficient(Variable(_n))) != 0){
         _breached=true;
         return true;
      }         
   }

   _breached=false;

   return false;
}


void TrRelImpl::substitute_updates(C_Polyhedron const & what, C_Polyhedron & result){
   Constraint_System const & cs= what.constraints();
   Constraint_System::const_iterator vi;

   PRECONDITION( result.space_dimension()== (unsigned) (_n+1), " Wrong space dimension in result");
   
   
   int i;
   for(vi= cs.begin(); vi != cs.end(); ++vi){
      Constraint const & cc = (*vi);
      Linear_Expression ln(0);
      for (i=0;i < _n+1; ++i){
         Linear_Expression & ll = (_dupdate.find(i))-> second;
         ln += cc.coefficient(Variable(i)) * ll;
      }
      if (cc.is_strict_inequality())
         result.add_constraint(ln > 0);
      else if (cc.is_nonstrict_inequality())
         result.add_constraint(ln >= 0);
      else if (cc.is_equality())
         result.add_constraint(ln ==0);
      else {

         //
         // this is unreachable
         //
         
         INVARIANT (false, " Unknown constraint type");
      }
   }

   return;
}


void TrRelImpl::dual_post( C_Polyhedron const & what, C_Polyhedron & result){
   //
   // take the polyhull of result with dguard.
   //

   if (what.is_universe())
      return;
   
   C_Polyhedron temp(what);

  
   // compute hull with _dguard
   
   //temp.poly_hull_assign(_dguard);

   Generator_System::const_iterator vi;

   for (vi= _guard_gens.begin(); vi!= _guard_gens.end(); ++vi){
      if (temp.relation_with((*vi))== Poly_Gen_Relation::subsumes())
         continue;
      temp.add_generator((*vi));
   }
   
   //
   // check if the guard has been breached
   // by examining temp
   //
   
   if (!guard_breached(temp)){
      return;
   }

   
   //
   // Now substitute the expressions on the constraints
   //

   substitute_updates(temp,result);
   
   return;
}

void TrRelImpl::post(C_Polyhedron const & what, C_Polyhedron & result) const {

   result=what;
   result.intersection_assign(_guard);
   result.add_space_dimensions_and_embed(_n);

   Variables_Set to_remove;
   // now add constraints for the updates
   for (int i=0;i<_n;++i){
      Linear_Expression ll(get_primal_update_expression(i));
      ll-=Variable(i+_n);
      result.add_constraint(ll==0);

      to_remove.insert(Variable(i));
   }

   result.remove_space_dimensions(to_remove);
   
}


void TrRelImpl::print_dual(ostream & out) const{

   var_info * fd1= _prel.get_dual_var_info();
   var_info * fd2= _postl.get_dual_var_info();

   out<<"Dual Guard [["<<endl;
   print_polyhedron(out,_dguard,fd1);
   out<<"]]"<<endl;

   out<<"Dual update relations [["<<endl;
   for (int i=0;i<=_n;++i){
         out<< fd1->get_name(i)<<" := ";
         print_lin_expression(out, (_dupdate.find(i))->second , fd2);
         out<<endl;
   }
   out <<"]]"<<endl;
   return;
}


Linear_Expression const & TrRelImpl::get_update_expression(int i) const{

   PRECONDITION( ( i >= 0 && i <= _n), " Index out of range ");
   ILEMap::const_iterator vi = _dupdate.find(i);
   INVARIANT( vi != _dupdate.end(), " Update relation not found ");

   return vi->second;
}


Linear_Expression const & TrRelImpl::get_primal_update_expression(int i) const{

   PRECONDITION( ( i >= 0 && i < _n), " Index out of range ");
   ILEMap::const_iterator vi = _update.find(i);
   INVARIANT( vi != _update.end(), " Update relation not found ");

   return vi->second;
}


void TrRelImpl::compute_pre_constraint(Constraint const & what, C_Polyhedron & result) const {
   Linear_Expression ll(0);
   int i;
   
   for (i =0 ; i < _n; ++i){
      ll += what.coefficient(Variable(i)) * get_primal_update_expression(i);
   }
   ll+= what.inhomogeneous_term();

   if (what.is_equality())
      result.add_constraint(ll==0);
   else if (what.is_nonstrict_inequality())
      result.add_constraint(ll >= 0);
   else if (what.is_strict_inequality())
      result.add_constraint(ll > 0);
}

void TrRelImpl::compute_pre_poly(C_Polyhedron const & post_poly, C_Polyhedron & result) const{
   
   if (post_poly.is_empty()) {
      result=C_Polyhedron(_n, EMPTY);
      return;
   }
   
   result=C_Polyhedron(_n, UNIVERSE);
   
   // just iterate through each constraint and compute a pre by adding substitutions
   Constraint_System cs = post_poly.constraints();
   Constraint_System::const_iterator vi;

   for (vi = cs.begin(); vi != cs.end(); ++vi){
      compute_pre_constraint((*vi), result);
   }
}
