
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

#ifndef D__TRRELATIONIMPL__H_
#define D__TRRELATIONIMPL__H_

#include <iostream>
#include <vector>
#include <map>
#include "Rational.h"
#include "var-info.h"
#include <ppl.hh>
#include "PolyUtils.h"
#include "TrLoc.h"

using namespace std;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;

//
// A simple transition relation with a guard
// and an update relation
//
extern bool dualize_guard;

typedef pair<int,Linear_Expression> ILEPair;
typedef map<int,Linear_Expression> ILEMap;

class TrRelImpl{

  private:

   int _n; // the number of variables
   var_info * _f;
   TrLoc const &  _prel;   
   TrLoc const &  _postl;
   
   C_Polyhedron _guard;
   C_Polyhedron _dguard; // the dual of the guard
   Generator_System _guard_gens;
   
   ILEMap _update; // the primal update map
   ILEMap _dupdate; // the dual update map
   bool _breached;
   bool _petri; // this transition is a petri net

   string _name;
   
   bool update_exists(int i) const;
   void initialize_dual_update();
   bool guard_breached(C_Polyhedron const & what);
   void substitute_updates(C_Polyhedron const & temp, C_Polyhedron & result);
   void compute_pre_constraint(Constraint const & what, C_Polyhedron & result) const;


   bool is_update_constant(int i);
   
  public:

   // constructor
   TrRelImpl(TrRelImpl const & what);

   TrRelImpl(TrLoc & prel, TrLoc & postl,
         C_Polyhedron const & g);

   // require an assignment operator for overloading
   
   // access the elements

   TrLoc const & get_preloc() const{
      return _prel;
   }

   TrLoc const & get_postloc() const{
      return _postl;
   }

   string const & get_preloc_name() const{
      return _prel.get_name();
   }

   string const & get_postloc_name() const{
      return _postl.get_name();
   }
   
   C_Polyhedron const & get_guard() const{
      return _guard;
   }
   
   C_Polyhedron const & get_dual_guard() const{
      return _dguard;
   }

   var_info * get_var_info() const{
      return _f;
      
   }
   
   int get_dimensions() const{
      return _n;
   }
   // add an update relation

   void add_update( int i, Linear_Expression const & what);
   void add_update( Variable v, Linear_Expression const & what){
      add_update((int) v.id(), what);
   }
  
   // Is this transition relation a petri net

   bool is_petri_net() const{
      return _petri;
   }
   

   //
   // for all those variables which do not
   // have updates, add the constant update
   //

   Linear_Expression const & get_update_expression(int i) const;
   
   void add_constant_updates();
   
   // routine to print the transition
   void print(ostream & out) const;

   // routine to compute the post condition

   void post(C_Polyhedron const & what, C_Polyhedron & result) const;

   // routine to compute the dual post

   void dual_post(C_Polyhedron const & what, C_Polyhedron & result);

   void print_dual(ostream & out) const;

   void compute_pre_poly(C_Polyhedron const & post_poly, C_Polyhedron & result) const;

   void set_name(string what){
      _name=what;
   }

   Linear_Expression const & get_primal_update_expression(int i) const;
      
   string const & get_name() const{
      return _name;
   }
   
   bool self_looping() const {
      return get_preloc_name()==get_postloc_name();
   }
   
};

// the print operator.

ostream & operator<<( ostream & out, TrRelImpl const & tt);


#endif
