
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */

            
/*           
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * License: Please read the included LICENSE file. 
 *          
 */                                     

#include "myassertions.h"
#include "TrSys.h"
#include "Timer.h"
#include "GenPoolMap.h"

extern long int prop_time, widening_time, total_time;

bool TrSys::f_check_duplicates(TrLoc const & loc)const {

   vector<TrLoc>::const_iterator vi;

   string const & w = loc.get_name();
   
   for (vi=_locs.begin(); vi != _locs.end(); ++vi){
      if (vi->get_name()==w)
         return true;
   }

   return false;
}


void TrSys::add_location(TrLoc const & loc){
   //
   // add the location loc
   //

   // first check if loc has been duplicated
   
   PRECONDITION(!f_check_duplicates(loc),
                "Trying to add two locations with the same name"<<loc.get_name()<<endl);

   // now push it into the vector
   
   
   _locs.push_back(loc);
}

void TrSys::add_transition(TrRel const & rel){
   //
   // add the transition relation rel
   //

   // now check if the pre and post locations of rel
   // are already in

   PRECONDITION( f_check_duplicates(rel.get_preloc()) ,
                 " Non existent prelocation "<< (rel.get_preloc()).get_name()<<endl
                 );

   PRECONDITION( f_check_duplicates(rel.get_postloc()) ,
                 " Non existent postlocation "<< (rel.get_postloc()).get_name()<<endl
                 );

   // now add it

   _rels.push_back(rel);

   return;
   
}

void TrSys::print(ostream & out) const{
   // print all the locations
   vector<TrLoc>::const_iterator vi;
   out<<"-----------------------------------------------------------"<<endl;
   out<<"System Locations"<<endl;
   for (vi=_locs.begin(); vi != _locs.end(); ++vi){
      out<<(*vi)<<endl;
   }

   out<<"System Transitions"<<endl;

   vector<TrRel>::const_iterator vj;
   for (vj=_rels.begin(); vj != _rels.end(); ++vj){
      out<<(*vj)<<endl;
   }
   out<<"-----------------------------------------------------------"<<endl;
}

ostream & operator<<(ostream & out, TrSys const & what){
   what.print(out);
   return out;
}

void TrSys::dual_post(IMap const & now, IMap & result){
   // set result to now
   result=now;

   // for each transition, compute the result of the post
   vector<TrRel>::iterator vi;
   
   for (vi= _rels.begin(); vi != _rels.end(); ++ vi){

      C_Polyhedron const & now_poly = now(vi->get_preloc());
      C_Polyhedron  & res_poly = result[vi->get_postloc()];

      vi->dual_post(now_poly,res_poly);
   }

   return;
}



void TrSys::compute_invariants(IMap & result, int prop_steps){
   dispatch(prop_steps,result);
}

void TrSys::dispatch(int k, IMap & result){
   //
   // dispatch according to the strategy
   //

   IMap now (_locs);

   Timer prop_timer;
   
   perform_initial_propagation(k, now);

   prop_timer.stop();
   prop_time=prop_timer.compute_time_elapsed();
   
   if (_strategy_id_ == 1){   
      narrow_until_termination_1(now);
   }

   if (_strategy_id_ == 2){
      narrow_until_termination_2(now);
   }
   
   now.primalize(result);
   return;
      
}


void TrSys::perform_initial_propagation( int k , IMap & result){

   PRECONDITION ( k > 0, " Asked to propagate for -ve number of steps");

   int i;

   IMap  next(_locs);
   dual_post(result,next);
   
   for (i=0;i<k;++i){
      result=next;
      cout<<" After "<<i<<" steps the map is :"<<endl;
      cout<<result<<endl;
      cout<<endl;
      dual_post(result,next);
   }
}

void TrSys::narrow_until_termination_1(IMap & now){

   //
   // narrow the map until termination
   //
 
   IMap next(now);
   do {
      now=next;
      
      
      Timer prop_timer;
      

      dual_post(now,next);

      prop_timer.stop();
      prop_time+=prop_timer.compute_time_elapsed();
 
      Timer widen_timer;
      
      now.narrow_with(next); 
      widen_timer.stop();
      widening_time+=widen_timer.compute_time_elapsed();
      
   } while(!next.contains(now));
   

}



void TrSys::narrow_until_termination_2(IMap & now){

   //
   // narrow the map until termination
   //

   //
   // create a map of generator pools
   //
   
   GenPoolMap genpools(now);
   IMap next(now);
   
   do {
      
      
      genpools.update_imap(now);

      // cout<<now<<endl;
      cout<<endl;

      
      Timer prop_timer;
      

      dual_post(now,next);

      prop_timer.stop();
      prop_time+=prop_timer.compute_time_elapsed();
 

      Timer widen_timer;
      genpools.narrowing_assign(next);
      widen_timer.stop();
      widening_time+=widen_timer.compute_time_elapsed();
      
      
   } while(!next.contains(now));
   

}
