
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include "funcs.h"
#include "cars-driver.h"


var_addrs vaddr;

// to access the variables

Variable v_x(int i){

   PRECONDITION( 0 <= i && i < ncars, \
                 " Dimension Mismatch");

   return Variable(vaddr.x[i]);

}


Variable v_v(int i){

   PRECONDITION( 0 <= i && i < ncars, \
                 " Dimension Mismatch");

   return Variable(vaddr.v[i]);

}


Variable v_a(int i){

   PRECONDITION( 0 <= i && i < ncars, \
                 " Dimension Mismatch");

   return Variable(vaddr.a[i]);

}

Variable v_r(int i){
   PRECONDITION( 1 <= i && i < ncars, \
                 " Dimension Mismatch");

   return Variable(vaddr.r[i]);

}


Variable v_t(){

   return Variable(vaddr.t);
}


Variable v_ac(){

   return Variable(vaddr.ac);
}


Variable v_br(){

   return Variable(vaddr.br);
}


/*
 * Form the system
 */

void form_var_info(){
   // first add the xi,vi and ai variables

   int i;
   int varcount=0;
   
   
   for (i=0;i<ncars;++i){
      
      string istr = int_to_str(i);

      string xi="x"+istr;
      vaddr.x[i]=varcount++;
      f->insert(xi.c_str());

      string vi="v"+istr;
      f->insert(vi.c_str());
      vaddr.v[i]=varcount++;

      string ai="a"+istr;
      f->insert(ai.c_str());
      vaddr.a[i]=varcount++;
      
      if (i >= 1){
         string ri="r"+istr;
         f->insert(ri.c_str());
         vaddr.r[i]=varcount++;
      }
         
   }

   vaddr.t=varcount++;
   f->insert("t");
   vaddr.ac=varcount++;
   f->insert("ac");
   vaddr.br= varcount++;
   f->insert("br");
   POSTCONDITION( varcount==dimension, " Something wrong!");
   
   //done
}



void form_invariant_polyhedron(C_Polyhedron & inv){

   PRECONDITION((int)inv.space_dimension()==dimension, \
                " Error: dimension mismatch");

   // now add the invariant

   // t >=0
   inv.add_constraint( v_t() >= 0);

   // ac = 2

   inv.add_constraint( v_ac()-2 == 0);

   // br = -2
   inv.add_constraint( v_br()+2 == 0);

   // now vi >=0 ac <= ai <= br, 0 <= ri <= 5

   for (int i=0;i<ncars;++i){
      // vi >= 0
      inv.add_constraint( v_v(i) >= 0);
      // vi <= 100
      //inv.add_constraint( v_v(i) <= 100);

      // ac >= ai >= br
      
      inv.add_constraint( v_a(i) <= v_ac());
      inv.add_constraint( v_a(i) >= v_br());

      // 0 <= ri <= 5 i>=1
      if (i >=1 ){
         inv.add_constraint(v_r(i) >= 0);
         inv.add_constraint(v_r(i) <= 5);
      }
   }
   
}

void form_initial_polyhedron(C_Polyhedron & theta){
   
   PRECONDITION((int)theta.space_dimension()==dimension, \
                " Error: dimension mismatch");

   int i;
   
   for(i=0;i<ncars;++i){
      // xi = 40 * (ncars -i)

      theta.add_constraint(v_x(i)== 40 * (ncars-i));

      // vi = 15 if (i=0) =5 if i=1, 0 o.w.
      if (i==0){
         theta.add_constraint(v_v(0)==15);
      } else if (i==1){
         theta.add_constraint(v_v(1)==5);
      } else {
         theta.add_constraint(v_v(i)==0);
      }

      // ai =0

      theta.add_constraint(v_a(i)==0);

      if (i>=1)
         theta.add_constraint(v_r(i)==0);
      
   }

   theta.add_constraint(v_ac()==2);
   theta.add_constraint(v_br()+2==0);

   theta.add_constraint(v_t()==0);

   // done
   
   
}

void form_system(){

   int i;
   
   // set the dimension
   // xi,vi,ai, for all cars
   // ri for every car other than the lead
   // time t, ac, br

   dimension= 4* ncars +2 ;

   // form the address structure
   
   vaddr.x = (int *) malloc(ncars * sizeof(int));
   vaddr.v = (int *) malloc(ncars * sizeof(int));
   vaddr.a = (int *) malloc(ncars * sizeof(int));
   vaddr.r = (int *) malloc(ncars * sizeof(int));

   
   // form the var_info
   form_var_info();

   // form the invariant polyhedron

   C_Polyhedron inv_poly(dimension,UNIVERSE);
   form_invariant_polyhedron(inv_poly);
 
   // form the initial polyhedron

   C_Polyhedron theta(dimension,UNIVERSE);
   form_initial_polyhedron(theta);

   

   //
   // Form the location
   //

   TrLoc trl(dimension, f, string("l0"), theta);
   vlocs->push_back(trl);

   //
   // form the transitions
   // 

   // type (a) -- time elapse transitions

   vector<bool> pt(ncars,false);

   // now use pt to form the time elapse transition

   do{
      add_time_elapse_transition(inv_poly, pt);
   } while (increment_pattern(pt));

   // now construct the transitions for each car.
   for (i=0;i<ncars;++i){
      construct_transition_for_car(inv_poly,i);
   }
   if (ncars <=3)
      make_system();

   // done
   
}

void construct_transition_for_car(C_Polyhedron  & inv, int i){
   TrLoc & l0= *(vlocs->begin());
   
   if (i==0){
      //
      // construct for the lead car
      // two transitions representing a non-deterministic
      // choice between acceleration and braking
      //

      C_Polyhedron g(inv);
      
      TrRel tra(l0,l0,g);
      TrRel trb(l0,l0,g);

      tra.add_update(v_a(0),v_br());
      trb.add_update(v_a(0),v_ac());
      tra.add_constant_updates();
      trb.add_constant_updates();
      vrels->push_back(tra);
      vrels->push_back(trb);
      
      

      return;
   }

   //else construct the three transitions

   C_Polyhedron g1(inv);

   g1.add_constraint(v_x(i-1) - v_x(i) <= 30);
   g1.add_constraint(v_x(i-1) - v_x(i) >= 0);

   TrRel tr1(l0,l0,g1 );
   tr1.add_update(v_a(i), v_br());
   tr1.add_update(v_r(i), Linear_Expression(0));

   tr1.add_constant_updates();

   vrels->push_back(tr1);

   
   C_Polyhedron g2(inv);

   g2.add_constraint(v_x(i-1) - v_x(i) >= 30);
   g2.add_constraint(v_x(i-1) - v_x(i) <= 50);

   TrRel tr2(l0,l0,g2 );
   tr2.add_update(v_a(i), Linear_Expression(0));
   tr2.add_update(v_r(i), Linear_Expression(0));

   tr2.add_constant_updates();

   vrels->push_back(tr2);

   
   
   C_Polyhedron g3(inv);

   g2.add_constraint(v_x(i-1) - v_x(i) >= 50);

   TrRel tr3(l0,l0,g3 );
   tr3.add_update(v_a(i), v_ac());
   tr3.add_update(v_r(i), Linear_Expression(0));

   tr3.add_constant_updates();

   vrels->push_back(tr3);


   //done
   

}

void add_time_elapse_transition(C_Polyhedron & inv, vector<bool> const & pattern){

   //
   // form the guard and update relations for the transition
   //

   TrLoc & l0= *(vlocs->begin());
   C_Polyhedron guard(inv);
   int i;
   
   for (i=1;i<ncars;++i){
      if (pattern[i]){

         // x{i-1} - x{i} >= 30

         guard.add_constraint(v_x(i-1)- v_x(i) >= 30);
         
         
      } else{

         // x{i-1} - x{i} <= 30
         // x{i-1} - x{i} >= 0
         
         guard.add_constraint(v_x(i-1)- v_x(i) <= 30);
         guard.add_constraint(v_x(i-1)- v_x(i) >= 0);
         
      }
   }
   
   TrRel trl(l0,l0,guard );

   // now add updates

   
   for(i=0;i<ncars;++i){

      // xi:= xi + vi + ai

      trl.add_update(v_x(i), v_x(i) + v_v(i) + v_a(i));

      // vi := vi + 2 * ai

      trl.add_update(v_v(i), v_v(i) + 2* v_a(i));

      // ai := ai

      // ri:= ri+1 if !pattern[i] and i >= 1
      if (i>=1 )
         if (!pattern[i]){
            trl.add_update(v_r(i), v_r(i)+1);
         }else{
            trl.add_update(v_r(i),Linear_Expression(0));
         }
      
   }

   trl.add_update(v_t(), v_t()+1);
   trl.add_constant_updates();

   vrels->push_back(trl);

}

bool increment_pattern(vector<bool> & pattern){

   int i;

   for (i = ncars-1; i >= 1 && pattern[i]; --i){
      pattern[i]=false;
   }
   if (i>=1){
      pattern[i]=true;
      return true;
   }

   return false;
      
}

int main(int argc, char * argv[]){

   parseinit(argc,argv);
   parse_command_line(argc,argv);

   
   form_system();

   if (ncars<=3){
      
      cerr<< (*global_system)<<endl;

   }
      //dispatch
   
   if (propagate)
      ch79_invariant_generation();
   else
      LP_invariant_generation();
   
   
   return 1;
   
    
}
