
/*
 * LpInv: Prototype Program Analyzer for Linear Systems using 
 *        Linear Programming. 
 * Author: Sriram Sankanarayanan <srirams@theory.stanford.edu> 
 * Copyright (c) 2005 Sriram Sankaranarayanan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *        
 */


#include "myassertions.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include "Timer.h"
#include "parse_globals.h"
#include "Eqn.h"
#include "EquationTemplate.h"
#include "LocTemplateMap.h"


using namespace std;

var_info * f;
vector<TrLoc> * vlocs;
vector<TrRel> * vrels;
int dimension;
int linenum;
int prop_steps;
int _max_evolutions;
bool perform_pre_optimization;
int _strategy_id_;
long int prop_time;
long int widening_time;
long int total_time;
long time_lp_invariants;
long time_map_setup;

int allowed_updates_;
int prop_depth;
int num_equations_orig;
int num_equations_surviving;
int num_equations_before_enrich;
bool dz_widen;

map<string,C_Polyhedron> * poly_var_map;

bool final_set;
C_Polyhedron * final_poly;
TrSys * global_system;

bool has_templates;
vector<TemplateBuilder> glb_templates;

bool colon;
bool propagate;
int zap_size;
bool dualize_guard;
bool lpinv;

int ncars;

/* Stats for CS04 technique */

cs_statistics cs_stats;
ch79_statistics ch79_stats;

ResourceMonitor mem_monitor;


bool evolution_criterion(int x){
   return (x <= allowed_updates_);
   
}




void parseinit( int argc, char * argv[]){
   prop_depth=3;
   has_templates=false;
   n_enrichments=2;
   f = new var_info();
   perform_pre_optimization=false;
   dz_widen=true;
   vlocs= new vector<TrLoc>();
   vrels= new vector<TrRel>();
   poly_var_map= new map<string,C_Polyhedron> ();
   _strategy_id_=2;
   final_set=false;
   linenum=0;
   time_map_setup=0;
   dimension=-1;
   allowed_updates_=3;
   prop_steps=3;
   _max_evolutions=0;
   prop_time=widening_time=total_time=0;
   time_lin_programming=0;
   n_lin_program_calls=0;
   n_simplex_iterations=0;
   time_lp_invariants=0;
   propagate=false;
   max_dimension_lp=-1;
   tot_dimension_lp=0;
   zap_size=2;
   ncars=3;
   lpinv=true;
}

void process_header(){
   dimension= f->get_dimension();
}


int find_variable(string * what){
   return f->search(what->c_str());
}


bool location_exists(string * what){
   vector<TrLoc>::iterator vi;
   for (vi= vlocs->begin(); vi != vlocs->end(); ++vi){
      if (vi->get_name()== (*what))
         return true;
   }
   return false;
}


TrLoc & location_reference(string * what){
   vector<TrLoc>::iterator vi;
   for (vi= vlocs->begin(); vi != vlocs->end(); ++vi){
      if (vi->get_name()== (*what))
         return (*vi);
   }

   UNREACHABLE( "Asked to find an unknown location "<<(*what) );
   return (*vi);
}

TrLoc & location_reference(){
   PRECONDITION( vlocs->size() == 1, " Wrong size of the location set to be calling this fn");

   return *(vlocs->begin());
}


void add_updates_to_transitions(TrRel & what, vector<ILEPair> * upds){
   vector<ILEPair>::iterator vi;

   for (vi=upds->begin(); vi != upds->end(); ++vi){
      what.add_update(vi->first, vi->second);
   }

   what.add_constant_updates();
}



int yyerror(){
   cerr<< " Error Line number "<<linenum<<endl;
   exit(1);
   return 0;
}

int yyerror(char * what){
   cerr<< what<< endl;
   return yyerror();
}


int yyerror(string  what){
   cerr<< what<< endl;
   return yyerror();
}


void make_system(){
   global_system = new TrSys(dimension, f);

   vector<TrLoc>::iterator vl;
   for (vl= vlocs->begin(); vl!= vlocs->end(); ++vl){
      global_system->add_location((*vl));
   }

   vector<TrRel>::iterator vtr;
   for (vtr=vrels->begin(); vtr!= vrels->end(); ++vtr){
      global_system->add_transition((*vtr));
   }
   

}

void make_system_and_propagate(){
   
   make_system();
   
   IMap resmap(*vlocs);
   Timer total_timer;
   global_system->compute_invariants(resmap, prop_steps);

   total_timer.stop();
   total_time=total_timer.compute_time_elapsed();
   
   resmap.print_using(std::cout, f);

   cout<<" Time for initial propagation (0.01S):"<<prop_time<<endl;
   cout<<" Time for Widening (0.01S):"<<widening_time<<endl;
   cout<<" Time total (0.01S):"<<total_time<<endl;
   
}

void LP_invariant_generation(){
   mem_monitor.take_a_reading();

#ifndef NDEBUG
   
   make_system();
   
   cout<< "The system read in "<<endl;
   
   cout << (*global_system)<<endl;
   cout << endl;
#endif
   
   
   mem_monitor.take_a_reading();
   
   Timer total_timer;
   Timer map_timer;
   cerr << " Forming the Initial Template Constraints....."<<endl;
   
   LocTemplateMap ltm(dimension, f, *vlocs, *vrels, glb_templates);
   map_timer.stop();
   time_map_setup= map_timer.compute_time_elapsed();
   mem_monitor.take_a_reading();
   
   cerr << " The initial TCM is: "<<endl;
   cerr << ltm<< endl;
   mem_monitor.take_a_reading();
   
   num_equations_orig=ltm.get_total_template_size();
   
   //   global_system->compute_lp_invariants();
   cerr<<" Computing Invariants..."<<endl;
   ltm.compute_invariants();
   mem_monitor.take_a_reading();
   time_lp_invariants= total_timer.compute_time_elapsed();
   num_equations_surviving=ltm.get_total_template_size();
   
   cerr<<" Invariant computation done."<<endl;
   print_stats_lp();
   IMap result((*vlocs));
   ltm.fill_imap(result);


   cerr << " Invariants computed "<<endl;
      result.print_using(cerr, f);
   cerr << " Checking .... ";
   result.check(*vrels);
   cerr << " Done "<<endl;

}

bool poly_var_exists(string * what){
   return (poly_var_map->find((*what))!= poly_var_map->end());
}

C_Polyhedron & get_polyhedron( string * what){
   PRECONDITION( poly_var_exists(what) , " Asked to retreive a polyhedron for a non-existent variable");

   map<string, C_Polyhedron>::iterator vi= poly_var_map->find((*what));

   return vi->second;
}

void add_poly_definition( string * name, C_Polyhedron * what){
   PRECONDITION( !poly_var_exists(name) , " Duplicate definition for polyhedron" << *name );

   poly_var_map->insert( pair<string,C_Polyhedron>(*name, *what));
   // done
}


void print_stats_lp(){
   
   //Basic problem information
   cerr << " Problem Dimensions:"<<dimension<<" #Locs:"<<vlocs->size()<< " #Trs : "<<vrels->size();
   // Information on Template size
   // compute total template size
   
   cerr<<" Total template size (original) "<< num_equations_before_enrich<<endl;
   cerr<<" After Enrichment"<< num_equations_orig<< " (Surviving) "<<num_equations_surviving<<endl;
   
   // Basic information on Complexity
  
   cerr << " Total time taken (0.01 s): "<< time_lp_invariants<<endl;
   cerr << " Time spent linear programming: "<< time_lin_programming<<endl;
   cerr << " Time taken to setup and enrich map:"<< time_map_setup<<endl;
   cerr << " # of calls to simplex: "<<n_lin_program_calls<<endl;
   cerr << " Average # of simplex iterations per call: "<< (float) n_simplex_iterations / (float) n_lin_program_calls <<endl;
   cerr << " Maximum dimensionality of LP problem:" << max_dimension_lp<<endl;
   cerr << " Avg. dimensionality of LP problem:" << (double) tot_dimension_lp/(double) n_lin_program_calls<<endl;

   
   cerr<<" Max. Resident Memory Size = "<<mem_monitor.get_maximum()<<endl;
}


void options_search_and_action(char * what){

   //
   // simple ... search if what is something
   // that I care about.. if so.. then act on it.
   //

   // - is optional before any option
   
   if (what[0]=='-') what=what+1;

   // Am I being aksed to run michael's code?

   if (strcmp(what,"dzoff")==0){
      dz_widen=false;
      return;
   }

  
   if (strcmp(what, "propagate")==0 ){
      colon=false;
      propagate=true;
      return;
   }

   //
   // Enrichment factor set?
   //
   if (what[6]=='='){

      what[6]='\0';

      if (strcmp(what,"enrich")==0){
         n_enrichments = atoi(what+7);
      }

   }

   if (what[7]=='='){
      what[7]='\0';

      if (strcmp(what,"zapsize")==0){
         zap_size = atoi(what+8);
      }
   }

   // Propagation depth

   if (what[5]=='='){
      what[5]='\0';
      
      if (strcmp(what,"depth")==0){
         prop_depth= atoi(what+6);
         cout << " Setting prop_depth="<<prop_depth<<endl;
         colon=true;
      }

      
      if (strcmp(what,"ncars")==0){
         ncars= atoi(what+6);
         cout << " Setting numcars="<<ncars<<endl;
      }
   }

   

   
   
}

void parse_command_line(int argc, char * argv[]){
   int j;
   for (j=1; j< argc;++j){
      options_search_and_action(argv[j]);
   }

}



void ch79_invariant_generation(){
   Timer total_timer;
   
   mem_monitor.take_a_reading();
   
   IMap m1((*vlocs),false),m2((*vlocs),false);
   
   for (int i=0;i<prop_depth;++i){
      m1.propagation_assign((*vrels),m2);
      mem_monitor.take_a_reading();
      m1=m2;
      
   }

  

   // widening begins
  
   
   m1.propagation_assign((*vrels),m2);
  
   ch79_stats.prop_time= total_timer.compute_time_elapsed();
    
   bool termination=false;
   
   while( ! termination){
      mem_monitor.take_a_reading();
      cerr<<"Starting Widening"<<endl;
      m2.widening_assign(m1);
      mem_monitor.take_a_reading();
      
      
      m1=m2;
      Timer prop_timer;
      
      m1.propagation_assign((*vrels),m2);
      mem_monitor.take_a_reading();
      
      cerr<<" Widened.."<<endl;
      // m2.print_using(cerr,f);
      
      ch79_stats.prop_time+= prop_timer.compute_time_elapsed();
      //   cerr<<"m2="<<endl;
      //  m2.print_using(cerr,f);
      
      termination= m1.contains(m2) ;
   }

   ch79_stats.total_time=total_timer.compute_time_elapsed();
   get_mem_usage();
   print_stats_ch79();

   cerr << " Invariants computed "<<endl;
   
   m1.print_using(cerr, f);
   cerr << " Checking .... ";
   m1.check(*vrels);
   cerr << " Done "<<endl;

  
   
}

void get_mem_usage(){

   /* struct rusage rlim;
   
   // ask for my resourse usage
   int ret=getrusage(RUSAGE_SELF,&rlim);
   if (ret !=0 ){
      cerr<<" Error getting memory usage "<<endl;
      
   }

   cs_stats.mem_maxrss = rlim.ru_maxrss;
   cs_stats.mem_idrss = rlim.ru_idrss;
   cs_stats.mem_isrss = rlim.ru_isrss;

   // done
   */
   
   mem_monitor.take_a_reading();
   cs_stats.mem_maxrss= mem_monitor.get_maximum();
   ch79_stats.mem_maxrss= mem_monitor.get_maximum();
   
}

void print_stats_cs(){



   cerr << " Problem Dimensions:"<<dimension<<" #Locs:"<<vlocs->size()<< " #Trs : "<<vrels->size();
   
   cerr<<" Dimension Zapping?";

   if (dz_widen){
      cerr<<" Yes."<<endl;
      cerr<< " Zap size = "<< zap_size<<endl;
   } else
      cerr<<" No. "<<endl;

   
   cerr<< " Propagation Depth = "<<prop_depth<<endl;
   cerr<<" Total Time = "<<cs_stats.total_time<<endl;
   
   cerr<<" Initial Propagation Time = "<<cs_stats.prop_time<<endl;
   cerr<<" # of eliminated multipliers after propagation ="<<cs_stats.mult_dims<<endl;
   cerr<<" Widening Time = "<< cs_stats.widen_time<<endl;
   cerr<<" LRS Generator Extraction Time="<<cs_stats.generator_extract_time<<endl<<endl;

   
   cerr<<" Max. Resident Memory Size = "<<cs_stats.mem_maxrss<<endl;


   
   

}


void print_stats_ch79(){



   cerr << " Problem Dimensions:"<<dimension<<" #Locs:"<<vlocs->size()<< " #Trs : "<<vrels->size();
      
   cerr<< " Propagation Depth = "<<prop_depth<<endl;
   cerr<<" Total Time = "<<ch79_stats.total_time<<endl;
   
   cerr<<" Propagation Time = "<<ch79_stats.prop_time<<endl;
  
   cerr<<" Max. Resident Memory Size = "<<ch79_stats.mem_maxrss<<endl;
   

}

